__version__ = "0.1.0-dev1"

from .diff.areaStatusEnum import AreaStatusEnum
from .diff.change import Change, ChangeList
from .diff.compair import ImxObjectCompare, ImxPropertyCompare
from .diff.diffStatusEnum import DiffStatusEnum
from .diff.imxDiff import ImxDiff
from .domain.area.areaClassifiedTypeEnum import AreaClassifiedTypeEnum
from .domain.area.areaClassifier import AreaClassifier
from .domain.area.imxProjectArea import ImxProjectArea
from .domain.imx import Imx
from .domain.models.imxEnums import ImxSituationsEnum
from .domain.models.imxGeometry import ImxGeometry
from .domain.models.imxMetadata import (
    ImxObjectMetadata,
    ProjectMetadata,
    ProjectNaiadeRemarks,
    SituationMetadata,
)
from .domain.models.imxPuic import ImxPuic, PuicConstructedByAttribute, PuicKeyTypeEnum
from .domain.models.ImxRailConnectionInfo import (
    RailConnectionInfo,
    RailConnectionInfoLine,
    RailConnectionInfoPoint,
    RailConnectionInfos,
)
from .domain.models.imxRelation import ImxReffedObjects
from .domain.models.imxSituations import ImxProject, ImxSituation
from .repo.imxRepo import SituationRepo
from .repo.tree.imxTopoTreeObject import ImxTopoObject
from .repo.tree.imxTreeObject import ImxObject
from .repo.tree.objectTree import ObjectTree
from .utils.shapely_geojson import (
    GeoJsonFeature,
    GeoJsonFeatureCollection,
    ShapleyGeojsonEncoder,
    dump,
    dumps,
)

__all__ = [
    "AreaStatusEnum",
    "ChangeList",
    "Change",
    "ImxPropertyCompare",
    "ImxObjectCompare",
    "DiffStatusEnum",
    "ImxDiff",
    "Imx",
    "ImxSituationsEnum",
    "ImxReffedObjects",
    "AreaClassifiedTypeEnum",
    "ImxSituation",
    "ObjectTree",
    "ImxGeometry",
    "ImxProject",
    "ProjectMetadata",
    "ImxObjectMetadata",
    "ProjectNaiadeRemarks",
    "ImxProjectArea",
    "ImxPuic",
    "PuicKeyTypeEnum",
    "PuicConstructedByAttribute",
    "SituationRepo",
    "SituationMetadata",
    "ImxObject",
    "ImxTopoObject",
    "RailConnectionInfos",
    "RailConnectionInfo",
    "RailConnectionInfoPoint",
    "RailConnectionInfoLine",
    "AreaClassifier",
    "GeoJsonFeature",
    "GeoJsonFeatureCollection",
    "ShapleyGeojsonEncoder",
    "dump",
    "dumps",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from imxInsights.diff.diffStatusEnum import DiffStatusEnum


@dataclass
class ChangeList:
    """
    A list of changes representing the difference between two objects.

    Args:
        status (DiffStatusEnum): A value from the `DiffStatus` enum indicating the status of the parent object.
        properties (list[Change]): A list of `Change` objects representing the changes made to the parent object's properties.

    """

    status: DiffStatusEnum
    properties: list[Change]

    def __post_init__(self):
        self._ensure_status_matches_props()

    def _ensure_status_matches_props(self):
        statuses = set((change.status for change in self.properties))
        if self.status == DiffStatusEnum.CREATED and statuses != {DiffStatusEnum.CREATED}:
            raise ValueError("Object is created but not all properties are")  # pragma: no cover
        if self.status == DiffStatusEnum.DELETED and statuses != {DiffStatusEnum.DELETED}:
            raise ValueError("Object is deleted but not all properties are")  # pragma: no cover
        if self.status not in [DiffStatusEnum.CREATED, DiffStatusEnum.DELETED] and any(status.is_changed() for status in statuses):
            self.status = DiffStatusEnum.UPDATED

    def to_dict(self, strip_attribute: bool = False, compact: Optional[bool] = True) -> dict[str, str]:
        """
        Returns a dictionary representing the `ChangeList` object.

        The keys of the dictionary are the property names, and the values are the changes.

        Args:
          strip_attribute: Whether to strip "@" from the beginning of the property names.
          compact: Boolean to return compact or old and new values.

        Returns:
          dict: A dictionary representing the `ChangeList` object.
        """
        # if compact:
        return {ChangeList._key_name(change.key, strip_attribute): change.display_compact for change in self.properties}
        # else:
        #     return {ChangeList._key_name(change.key, strip_attribute): {"old": change.a, "new": change.b} for change in self.props}

    @staticmethod
    def _key_name(key: str, strip_attribute: bool) -> str:
        if strip_attribute:
            return key.replace("@", "")
        return key


@dataclass
class Change:
    """
    Represents a change in a property of an object.

    Args:
        key (str): The property key.
        status (DiffStatusEnum): The status of the change.
        a (Optional[str]): The old value of the property.
        b (Optional[str]): The new value of the property.

    """

    key: str
    status: DiffStatusEnum
    a: Optional[str]
    b: Optional[str]

    def __post_init__(self):
        if self.status == DiffStatusEnum.NOT_APPLICABLE:
            raise ValueError(f"Cannot instantiate a Change with {self.status}")  # pragma: no cover

    def __repr__(self) -> str:
        return self.display_compact

    def __str__(self) -> str:
        return self.display_compact

    def has_location(self) -> bool:
        """
        Determines if the property key is a location by checking if it matches certain gml patterns.

        Returns:
            bool: True if the property key is a location, False otherwise.
        """
        return True if self.key in [".coordinates", ".Point.", ".LineString."] else False
        # return ".coordinates" in self.key or ".Point." in self.key or ".LineString." in self.key

    @property
    def display_compact(self) -> str:
        """
        Returns a formatted string representing the change.

        The format depends on the status of the change.

        - If the status is DiffStatusEnum.NO_CHANGE, it returns the new value.
        - If the status is DiffStatusEnum.CREATED, it returns the new value with a "++" prefix.
        - If the status is DiffStatusEnum.DELETED, it returns the old value with a "--" prefix.
        - If the status is DiffStatusEnum.UPDATED, it returns the old value followed by "->" and the new value.
        - If the status is DiffStatusEnum.UPGRADED, it returns the old value followed by "=>" and the new value.
        - If the status is DiffStatusEnum.NOT_APPLICABLE, it returns "N/A".

        Returns:
            str: A formatted string representing the change.
        """
        if self.key == "puic":
            return self.b if self.b is not None else f"{self.a}"

        match self.status.value:
            case DiffStatusEnum.NO_CHANGE.value:
                return f"{self.b}"
            case DiffStatusEnum.CREATED.value:
                return f"++ {self.b}"
            case DiffStatusEnum.DELETED.value:
                return f"-- {self.a}"
            case DiffStatusEnum.UPDATED.value:
                return f"{self.a} -> {self.b}"
            case DiffStatusEnum.UPGRADED.value:
                return f"{self.a} => {self.b}"
            case DiffStatusEnum.NOT_APPLICABLE.value:
                return "N/A"
        return "N/A"



import asyncio

from arcGisFeatureCash import ArcGisFeatureService
from shapely import Point


async def get_feature_service_km_async():
    url = "https://mapservices.prorail.nl/arcgis/rest/services/Referentiesysteem_004/FeatureServer"
    return await ArcGisFeatureService.factory(url)


async def get_feature_service_tekeningen_async():
    url = "https://maps.prorail.nl/arcgis/rest/services/Tekeningen_schematisch/FeatureServer"
    return await ArcGisFeatureService.factory(url)


async def get_all_gis_features_async():
    # todo: make jobs list and await that one to gain some speed
    km_features = await get_feature_service_km_async()
    tekening_features = await get_feature_service_tekeningen_async()
    return km_features, tekening_features


class GisRailDataLayer:
    def __init__(self, feature_cash_km, feature_cash_schema_tek):
        self._feature_cash_km: ArcGisFeatureService = feature_cash_km
        self._feature_cash_schema_tek: ArcGisFeatureService = feature_cash_schema_tek

        self.km_vlak = [_ for _ in self._feature_cash_km.feature_service_layers if _.dataset == "Kilometerlintvlak"][0]
        self.km_lint = [_ for _ in self._feature_cash_km.feature_service_layers if _.dataset == "Kilometerlint"][0]
        self.OBE_contour = [_ for _ in self._feature_cash_schema_tek.feature_service_layers if _.dataset == "OBE blad"][0]

    def get_km_linear_referencing(self, x: float, y: float) -> str:
        def round_floor_to_3_decimals(num):
            return (num * 1000) // 1 / 1000.0

        point = Point(x, y)
        vlak_matches = [item for item in self.km_vlak.features if point.within(item.geometry)]
        response = []
        for item in vlak_matches:
            lint_name = item.attributes.get_value("KMLINT")
            lint_match = [item for item in self.km_lint.features if item.attributes.get_value("NAAM") == lint_name][0]
            projected_point = lint_match.measure_geometry.interpolate(lint_match.measure_geometry.project(point))
            response.append(f"{lint_name}: km {round_floor_to_3_decimals(projected_point.z)}")

        return "".join(response) if len(response) == 1 else "\n".join(response)

    def get_obe_contours(self):
        pass


def get_gis_rail_data_layer() -> GisRailDataLayer:
    km_features, schema_tek_features = asyncio.run(get_all_gis_features_async())
    return GisRailDataLayer(km_features, schema_tek_features)



import re


def write_situation_info(worksheet, situation, start_row: int, suffix=""):
    if suffix != "":
        suffix = f" {suffix}"

    worksheet.write(start_row + 0, 0, f"file{suffix}")
    worksheet.write(start_row + 0, 1, situation.file_path.name)

    worksheet.write(start_row + 1, 0, f"imx version{suffix}")
    worksheet.write(start_row + 1, 1, situation.imx_version)

    worksheet.write(start_row + 2, 0, f"sha256{suffix}")
    worksheet.write(start_row + 2, 1, situation.file_hash)

    worksheet.write(start_row + 3, 0, f"situation{suffix}")
    worksheet.write(start_row + 3, 1, situation.situation_type)

    worksheet.write(start_row + 4, 0, f"perspective_date{suffix}")
    worksheet.write(start_row + 4, 1, situation.perspective_date)

    worksheet.write(start_row + 5, 0, f"reference_date{suffix}")
    worksheet.write(start_row + 5, 1, situation.reference_date)


def sorted_path_methode(value):
    parts = value.split(".")
    sorted_parts = []

    for part in parts:
        num_part = re.search(r"\d+", part)
        if num_part:
            sorted_parts.append((int(num_part.group()), part))
        else:
            sorted_parts.append((-1, part))  # Use a negative value to prioritize non-numeric parts

    return sorted_parts


def sort_pandas_dataframe_columns(df, column_order_pre_fix_list, column_order_suffix_list=None):
    if column_order_suffix_list is None:
        column_order_suffix_list = []

    columns_in_pre_fix_list = [col for col in column_order_pre_fix_list if col in df.columns]
    columns_in_sur_fix_list = [col for col in column_order_suffix_list if col in df.columns]

    columns_not_in_lists = [col for col in df.columns if col not in column_order_pre_fix_list and col not in columns_in_sur_fix_list]
    columns_not_in_lists = sorted(columns_not_in_lists, key=sorted_path_methode)
    ordered_columns = columns_in_pre_fix_list + columns_not_in_lists + columns_in_sur_fix_list

    return df[ordered_columns]



import xml.etree.ElementTree as ET
from dataclasses import dataclass
from enum import Enum
from typing import Optional

import matplotlib.pyplot as plt
from shapely import Point
from shapely.geometry import LineString
from shapely.ops import linemerge, unary_union

from imxInsights.utils.shapely_helpers import (
    extend_line,
    point_relative_to_linestring,
    reproject_point_to_line,
)

IMX_NS = {"": "http://www.prorail.nl/IMSpoor", "gml": "http://www.opengis.net/gml"}  # Global namespace dict


@dataclass
class ImxGraphPassage:
    puic: str
    name: str
    side_tag: str
    geometry: LineString
    speed: float | None = None


@dataclass(unsafe_hash=True)
class ImxGraphSwitchMechanism:
    puic: str
    name: str
    geometry: Point
    geometry_reprojected: Optional[Point] = None
    left_or_right_line: Optional[LineString] = None


class ImxGraphSwitchPositionEnum(Enum):
    LEFT = "left"
    RIGHT = "right"


@dataclass(unsafe_hash=True)
class ImxGraphSwitchPosition:
    switch_mechanism: ImxGraphSwitchMechanism
    position: ImxGraphSwitchPositionEnum


@dataclass
class ImxGraphJunctionPath:
    passages: list[ImxGraphPassage]
    geometry: LineString
    speed: float
    start: ImxGraphPassage
    end: ImxGraphPassage
    switches_in_path: list[ImxGraphSwitchPosition]


@dataclass
class ImxGraphPassageSwitchMechanismDistance:
    passage: str
    switch_mechanism: str
    distance: float


class ImxGraphJunction:
    def __init__(self, xml_data):
        self._r = ET.fromstring(xml_data)

        self.passages = self._parse_passages()
        self._st_passage = self._find_passage_by_name("S-T")
        self._pq_passage = self._find_passage_by_name("P-Q")
        self._center_line = self._calculate_center_line()

        self.switch_mechanisms = self._parse_switch_mechanisms()
        self._reproject_switch_mechanisms()
        self._passage_switch_distance_map = self._map_passages_to_switch_mechanisms()
        self._assign_switch_lines()

    @staticmethod
    def from_element(junction_element: ET.Element) -> "ImxGraphJunction":
        return ImxGraphJunction(ET.tostring(junction_element))

    @staticmethod
    def _parse_coordinates_linestring(coordinates_str: str) -> list[tuple[float, ...]]:
        return [tuple(map(float, coord.split(",")[:2])) for coord in coordinates_str.split()]

    def _parse_coordinates_point(self, coordinate_str: str) -> Point:
        return Point(self._parse_coordinates_linestring(coordinate_str)[0])

    def _parse_passages(self) -> dict[str, ImxGraphPassage]:
        return {
            passage.get("puic"): ImxGraphPassage(
                puic=passage.get("puic"),
                name=passage.get("name"),
                side_tag=passage.get("sideTag"),
                geometry=LineString(self._parse_coordinates_linestring(passage.find(".//gml:coordinates", IMX_NS).text)),
                speed=passage.get("passageSpeed", None),
            )
            for passage in self._r.findall(".//Passage", IMX_NS)
        }

    def _find_passage_by_name(self, name: str) -> Optional[str]:
        return next((puic for puic, passage in self.passages.items() if passage.name == name), None)

    def _parse_switch_mechanisms(self) -> dict[str, ImxGraphSwitchMechanism]:
        return {
            switch.get("puic"): ImxGraphSwitchMechanism(
                puic=switch.get("puic"),
                name=switch.get("name"),
                geometry=self._parse_coordinates_point(switch.find(".//gml:coordinates", IMX_NS).text),
            )
            for switch in self._r.findall(".//SwitchMechanism", IMX_NS)
        }

    def _calculate_center_line(self) -> LineString:
        passages_of_interest = {_.name: _ for puic, _ in self.passages.items() if _.name in ["R", "L", "V"]}
        if len(passages_of_interest.keys()) == 3:
            midpoints = {name: passage.geometry.centroid for name, passage in passages_of_interest.items()}
            left_midpoint = midpoints["V"].centroid
            right_midpoint = LineString([midpoints["L"], midpoints["R"]]).centroid

        elif len(self.passages.keys()) != 0:
            passages_of_interest = {_.name: _ for puic, _ in self.passages.items() if _.name in ["P", "T", "Q", "S"]}

            midpoints = {name: passage.geometry.centroid for name, passage in passages_of_interest.items()}
            left_midpoint = LineString([midpoints["P"], midpoints["T"]]).centroid
            right_midpoint = LineString([midpoints["Q"], midpoints["S"]]).centroid

        else:
            return LineString()

        center_point = LineString([left_midpoint, right_midpoint]).centroid
        center_line = LineString([left_midpoint, center_point, right_midpoint])

        return center_line

    def _reproject_switch_mechanisms(self) -> None:
        for switch in self.switch_mechanisms.values():
            switch.geometry_reprojected = reproject_point_to_line(switch.geometry, self._center_line)

    def _map_passages_to_switch_mechanisms(self) -> dict[str, Optional[ImxGraphPassageSwitchMechanismDistance]]:
        passage_switch_distance_map = {}

        for passage_id, passage_geom in self.passages.items():
            closest_switch = None
            min_distance = 0.0

            # the controlled switch mech is the one with the most distance to the passage
            for switch_id, switch_geom in self.switch_mechanisms.items():
                distance = passage_geom.geometry.distance(switch_geom.geometry_reprojected)
                if distance > min_distance:
                    min_distance = distance
                    closest_switch = (switch_id, distance)
            if len(self.switch_mechanisms.keys()) != 0:
                passage_switch_distance_map[passage_id] = ImxGraphPassageSwitchMechanismDistance(passage_id, *closest_switch)

        all_mapped = all(mapping is not None for mapping in passage_switch_distance_map.values())
        all_mapped_reversed = all(mapping is None for mapping in passage_switch_distance_map.values())
        if not all_mapped and not all_mapped_reversed:
            raise ValueError("Some passages could not be mapped to a switch mechanism.")

        return passage_switch_distance_map

    def _assign_switch_lines(self):
        for passage_id, switch_info in self._passage_switch_distance_map.items():
            if switch_info is None or switch_info.switch_mechanism is None:
                continue

            switch_mech_id = switch_info.switch_mechanism
            switch_mech = self.switch_mechanisms[switch_mech_id]
            passage = self.passages[passage_id]

            if not switch_mech.left_or_right_line:
                projected_passage_point = self._center_line.interpolate(self._center_line.project(passage.geometry.centroid))

                switch_mech_line = extend_line(switch_mech.geometry_reprojected, projected_passage_point, extension_meters=25, extend_from="end")

                switch_mech.left_or_right_line = switch_mech_line

    def _passage_left_or_right_of_switch(self, switch_mech: ImxGraphSwitchMechanism, passage: ImxGraphPassage) -> str:
        projected_passage_point = self._center_line.interpolate(self._center_line.project(passage.geometry.centroid))
        extended_line_from_switch = extend_line(switch_mech.geometry_reprojected, projected_passage_point, extension_meters=10, extend_from="end")
        is_passage_right_of_switch = point_relative_to_linestring(extended_line_from_switch, passage.geometry.centroid)
        return is_passage_right_of_switch

    @staticmethod
    def _get_speed_in_path(passages: list[ImxGraphPassage]) -> float:
        return float("inf") if all(value.speed is None for value in passages) else min([_.speed for _ in passages])

    @staticmethod
    def _validate_path(passages: list[ImxGraphPassage]) -> bool:
        if any(passage is None for passage in passages):
            return False

        geometries = [passage.geometry for passage in passages if passage]
        merged = linemerge(unary_union(geometries))
        return merged.is_valid and merged.geom_type == "LineString"

    def _get_path_geometry(self, path_puics: list[str]) -> LineString:
        start_passage = self.passages.get(path_puics[0])
        end_passage = self.passages.get(path_puics[-1])

        side_tag_insertions = {
            ("T", "S"): self._st_passage,
            ("S", "T"): self._st_passage,
            ("P", "Q"): self._pq_passage,
            ("Q", "P"): self._pq_passage,
        }
        side_tag_pair = (start_passage.side_tag, end_passage.side_tag)
        if side_tag_pair in side_tag_insertions and side_tag_insertions[side_tag_pair] is not None:
            path_puics = [path_puics[0], side_tag_insertions[side_tag_pair], path_puics[-1]]

        return linemerge([self.passages[puic].geometry for puic in path_puics if puic in self.passages])

    def get_path(self, path_puics: list[str]) -> ImxGraphJunctionPath:
        passages = [self.passages.get(puic) for puic in path_puics]
        if not self._validate_path(passages):
            raise ValueError("path is not valid")

        min_speed = self._get_speed_in_path(passages)
        geometry = self._get_path_geometry(path_puics)
        start = self.passages.get(path_puics[0])
        end = self.passages.get(path_puics[-1])

        switches_positions_in_path = []
        if len(self._passage_switch_distance_map.keys()) == 0:
            switches_positions_in_path = []
        else:
            for puic in path_puics:
                passage_switch_distance_map = self._passage_switch_distance_map[puic]

                switch_mech = self.switch_mechanisms[passage_switch_distance_map.switch_mechanism]

                passage = self.passages[puic]

                if passage.side_tag != "V":
                    left_or_right = self._passage_left_or_right_of_switch(switch_mech, passage)
                    switches_positions_in_path.append(ImxGraphSwitchPosition(switch_mech, ImxGraphSwitchPositionEnum[left_or_right.upper()]))

            switches_positions_in_path = list(set(switches_positions_in_path))

        return ImxGraphJunctionPath(
            passages=passages, speed=min_speed, geometry=geometry, start=start, end=end, switches_in_path=switches_positions_in_path
        )

    ####################################################################################################

    def plot_path_with_elements(self, path_geometries=None, path_title=""):
        fig, ax = plt.subplots()

        def plot_line(x, y, label, style_dict, with_markers=True):
            # Use the label only for the first call to this function
            ax.plot(x, y, label=label, **style_dict)
            if with_markers:
                ax.text((x[0] + x[-1]) / 2, (y[0] + y[-1]) / 2, label, fontsize=9, ha="center")

        # Plot passages
        passage_style = {"marker": "o", "linestyle": "-", "markersize": 5, "linewidth": 3}
        for passage in self.passages.values():
            x, y = passage.geometry.xy
            plot_line(x, y, passage.name, passage_style)

        # Plot switch mechanisms and reprojected switch mechanisms
        switch_colors = ["red", "orange"]
        for idx, switch in enumerate(self.switch_mechanisms.values()):
            # Plot left or right line for switch mechanism
            if switch.left_or_right_line:
                x, y = switch.left_or_right_line.xy

                switch_mechanism_line_style = {
                    "marker": "o",
                    "linestyle": "--",
                    "markersize": 5,
                    "linewidth": 5,
                    "color": switch_colors[idx % len(switch_colors)],
                }
                plot_line(x, y, f"{switch.name} switch mechanism line", switch_mechanism_line_style)
                ax.plot(x, y, linestyle="--", linewidth=5, color=switch_colors[idx % len(switch_colors)])

            # Original and reprojected switch positions with unique labels
            for geom, geom_label in zip([switch.geometry, switch.geometry_reprojected], ["", "\n(reprojected)"]):
                ax.plot(geom.x, geom.y, "^", markersize=10, color=switch_colors[idx % len(switch_colors)])
                ax.text(geom.x, geom.y, f"{switch.name}{geom_label}", fontsize=9, color="red", ha="center")

        # Plot center line

        ax.plot(*self._center_line.xy, label="Center Line", marker="", linestyle="--", linewidth=5, color="green")

        # Plot path
        path_style = {"marker": ">", "linestyle": "--", "linewidth": 6, "color": "magenta"}

        if path_geometries:
            x, y = path_geometries.xy

            for i in range(len(x) - 1):
                dx = x[i + 1] - x[i]
                dy = y[i + 1] - y[i]
                plt.arrow(x[i], y[i], dx, dy, head_width=1, head_length=1, fc="magenta", ec="magenta")

            plot_line(x, y, "Path", path_style)

        # Final plot adjustments
        ax.set_aspect("equal", adjustable="datalim")
        plt.xlabel("X coordinate")
        plt.ylabel("Y coordinate")
        plt.title(f"Graph junction {path_title}")
        plt.legend(loc="upper left", fontsize="small")
        plt.show()



from __future__ import annotations

from enum import Enum


class ImxSituationsEnum(str, Enum):
    """
    Valid situations in a imx project file.

    Attributes:
        InitialSituation (str): The initial situation in a imx project file.
        NewSituation (str): The new situation in a imx project file.
        Situation (str): A situation in a imx situation file.
    """

    InitialSituation = "InitialSituation"
    NewSituation = "NewSituation"
    Situation = "Situation"


class AreaNameEnum(Enum):
    """
    Valid imx areas in a imx project file.

    Attributes:
        CONTEXT: ContextArea.....
        USER: UserArea.....
        WORK: WorkArea.....
    """

    CONTEXT = "ContextArea"
    USER = "UserArea"
    WORK = "WorkArea"



from __future__ import annotations

from pathlib import Path
from typing import Dict, List, Optional

import pandas as pd
from lxml import etree as ET
from shapely import GeometryCollection, LineString, Point
from shapely.geometry.base import BaseGeometry
from shapely.ops import linemerge

from imxInsights.domain.models.imxEnums import AreaNameEnum
from imxInsights.domain.models.imxGeometry import ImxGeometry

# from imxInsights.domain.models.ImxRailConnectionInfo import (
#     RailConnectionInfoLine,
#     RailConnectionInfoPoint,
# )
from imxInsights.domain.models.imxRelation import ImxRelation
from imxInsights.graph.imxGraph import ImxGraph
from imxInsights.graph.imxGraphBuilder import ImxGraphBuilder
from imxInsights.graph.queries.sectionGeometryQuery import SectionGeometryGraphQuery
from imxInsights.repo.tree.imxTreeObject import ImxObject
from imxInsights.repo.tree.objectTree import ObjectTree
from imxInsights.utils.helpers import hash_sha256
from imxInsights.utils.log import logger
from imxInsights.utils.shapely_geojson import GeoJsonFeature, GeoJsonFeatureCollection
from imxInsights.utils.shapely_helpers import (
    ShapelyTransform,
    gml_linestring_to_shapely,
    gml_point_to_shapely,
    gml_polygon_to_shapely,
    reverse_line,
)
from imxInsights.utils.xml_helpers import (
    NS_GML,
    NS_IMSPOOR,
    XmlFile,
    find_coordinates,
    trim_tag,
)

# from shapelyM import MeasureLineString


class SituationRepo:
    """
    A repository of all imx value objects.

    ***Class is responsible for parsing a IMX situation xml element into value objects.***

    !!! danger "Value objects"
        ***ImxInsights is not scoped to edit an IMX file therefor we cant add or change value objects in the repository.***

    !!! info "Value objects"
        ***Every object with a puic attribute is a value object.*** Some objects do not have a puic attribute but are objects of intrest we can
        specify a custom puic on specific imx types.

        *The custom puic configuration can be found in the custom_puic_config.yaml.*

    !!! info "Value object geometry"
        Value object can have geometry specified in a gml node. If no gml node is present we can try to construct geometry:

        - RailConnection geometry is constructed by From- and ToNode attributes in the MicroLink.
        - TrackFragments will be constructed by from- and toMeasures, this results in LineString geometry.
        - Demarcation marker will be projected and results in (Point) geometry.

    ??? tip "Object path vs object type"
        Some IMX object are sub objects of many IMX objects types, by using path instead of imx types we can get only those of interest.

        As an example: we want to get all switch mechanisms of single switches, if we use the imx type Lock we get all Locks. Instead, use the
        Path SingleSwitch.SwitchMechanism to get only the locks that are sub objects of SingleSwitch.

    Attributes:
        element (ET.Element): Exml element of the situation.
        tree (ObjectTree): the tree containing all value objects.

    """

    def __init__(self, element: ET.Element, xml_file: XmlFile, add_graph: bool = False):
        assert element.tag.endswith("Situation"), "Element should be a imx situation"
        self._xml_file = xml_file
        self.imx_version = self._xml_file.root.find("[@imxVersion]").attrib["imxVersion"]
        self.file_path: Path = self._xml_file.path
        self.file_hash: str = hash_sha256(self._xml_file.path)
        self._element: ET.Element = element
        self.situation_type: str = trim_tag(self.element.tag)
        self._tree: ObjectTree = ObjectTree(self._element)
        self._link_reffed_objects()
        self._parse_geometry()
        self._construct_rail_connections_geometry()
        # self._set_azimuth_from_functional_direction()
        self._construct_track_fragment_geometry()
        self._construct_demarcation_markers()
        self._link_rail_connection_info()
        self._register_children()

        if add_graph:
            self.graph: ImxGraph = ImxGraphBuilder(self).build_graph()
            self._create_section_geometry()

    @property
    def reference_date(self) -> str:
        """
        Returns the external project reference as given in the project metadata, *read only property*.

        !!! Info
            The reference date is the date that the data is valid, it could be historic or feature situation.

        """
        return str(self._element.get("referenceDate"))

    @property
    def perspective_date(self) -> str:
        """
        Returns the perspective date as given in the situation metadata, *read only property*.

        !!! info
            The perspective date is the date the imx file is created and can be used to determinate if the data baseline is still up to date.

        """
        return str(self._element.get("perspectiveDate"))

    @property
    def element(self) -> ET.Element:
        return self._element

    @property
    def tree(self) -> ObjectTree:
        return self._tree

    def _parse_geometry(self):
        # todo: add parse azimuth as property
        for entity in self._tree.objects():
            location_node = entity.element.findall(f"{{{NS_IMSPOOR}}}Location")
            if len(location_node) == 0:
                entity.geometry = ImxGeometry(GeometryCollection())
            else:
                geographic_location = location_node[0].find(f"{{{NS_IMSPOOR}}}GeographicLocation")
                assert len(location_node) == 1, "main object should not have multiple locations"
                coordinates_element = location_node[0].findall(f".//{{{NS_GML}}}coordinates")
                coordinates = find_coordinates(coordinates_element[0])
                geometry = self._parse_coordinates(coordinates[0])
                azimuth = geographic_location.get("azimuth")
                if azimuth is not None:
                    azimuth = float(azimuth)
                entity.geometry = ImxGeometry(geometry, azimuth=azimuth)

    @staticmethod
    def _parse_coordinates(coordinates: ET.Element) -> BaseGeometry:
        assert coordinates.tag.endswith("coordinates"), "Please provide <gml:coordinates/> element"
        parent_tag = trim_tag(coordinates.getparent().tag)
        match parent_tag:
            case "Point":
                return gml_point_to_shapely(coordinates.text)
            case "LineString":
                return gml_linestring_to_shapely(coordinates.text)
            case "LinearRing":
                return gml_polygon_to_shapely(coordinates.text)
            case _:
                raise ValueError("Cannot parse coordinates for element")

    def _construct_rail_connections_geometry(self) -> None:
        rail_connections = self.get_by_types(["RailConnection"])
        for rail_connection in rail_connections:
            if rail_connection.puic == "07e61000-dc0c-4ff7-97b1-2201b37d6e8e":
                print()

            if "trackRef" in rail_connection.element.attrib:
                track_ref = rail_connection.element.attrib["trackRef"]
            else:
                track_ref = None

            if "passageRefs" in rail_connection.element.attrib:
                passage_refs = rail_connection.element.attrib["passageRefs"].split(" ")
            else:
                passage_refs = []

            if len(passage_refs) == 0:
                passage_refs = rail_connection.element.findall(".//{*}PassageRefs")
                if len(passage_refs) == 0:
                    passage_refs = []
                else:
                    if passage_refs[0].text is None:
                        passage_refs = []
                    elif " " in passage_refs[0].text:
                        passage_refs = passage_refs[0].text.split(" ")
                    else:
                        passage_refs = [passage_refs[0].text]

            geometries = []
            if len(passage_refs) != 0:
                for passage_ref in passage_refs:
                    passage_obj = self.tree.find(passage_ref)
                    if passage_obj is None:
                        raise ValueError(f"RailConnection {rail_connection.puic}: passageRef {passage_ref} not present in dataset")
                    geometries.append(passage_obj.shapely)

            if track_ref is not None:
                track_obj = self.tree.find(track_ref)
                if track_obj is None:
                    raise ValueError(f"RailConnection {rail_connection.puic}: trackRef {track_ref} not present in dataset")
                geometries.append(track_obj.shapely)

            line_geometry = linemerge(geometries)
            # todo: make it accept bad constructed geometry....?
            #  - seems z is not issue on merging line string.... z is only along the ride...
            assert isinstance(
                line_geometry, LineString
            ), f"RailConnection {rail_connection.puic} constructed geometry should be a single polyline geometry"

            logger.info(f'RailConnection {rail_connection.puic} geometry constructed from track: {track_ref} and passages:{", ".join(passage_refs)}')

            if rail_connection.topo_object is not None:
                from_node_point = self.tree.find(rail_connection.topo_object.element.findall(".//{*}FromMicroNode")[0].attrib["nodeRef"])
                if from_node_point is not None:
                    from_node_point = from_node_point.shapely

                to_node_point = self.tree.find(rail_connection.topo_object.element.findall(".//{*}ToMicroNode")[0].attrib["nodeRef"])
                if to_node_point is not None:
                    to_node_point = to_node_point.shapely

            first_coord_distance_from_from_node = Point(line_geometry.coords[0]).distance(from_node_point)
            first_coord_distance_from_to_node = Point(line_geometry.coords[0]).distance(to_node_point)

            if first_coord_distance_from_to_node < first_coord_distance_from_from_node:
                logger.debug(f'RailConnection {rail_connection.puic} "FromNode" seems to be on end of the polyline: reversed geometry')
                line_geometry = reverse_line(line_geometry)

            rail_connection.geometry = ImxGeometry(line_geometry)

    def _construct_track_fragment_geometry(self):
        for item in self.get_all():
            if item.track_fragments is not None:
                for rail_info in item.track_fragments.rail_infos:
                    if hasattr(rail_info, "from_measure") and hasattr(rail_info, "to_measure"):
                        rail_con = self.tree.find(rail_info.ref)

                        if not rail_con:
                            logger.error(f"{item.tag} {item.puic} doest not have valid railRef {rail_info.ref} cant create geometry")

                        elif rail_info.from_measure == rail_info.to_measure:
                            logger.warning(
                                f"{item.tag} {item.puic} doest not have valid TrackFragments (same from and to) railRef {rail_info.ref} cant create geometry"  # noqa: E501
                            )

                        else:
                            if rail_info.from_measure < rail_info.to_measure:
                                projection = rail_con.geometry.shapelyM.cut_profile(rail_info.from_measure, rail_info.to_measure)
                            else:
                                projection = rail_con.geometry.shapelyM.cut_profile(rail_info.to_measure, rail_info.from_measure)

                            if not projection.result:
                                logger.error(f"{item.tag} {item.puic} doest not have valid line cut result, cant create geometry")
                            else:
                                rail_info.geometry = projection.result.shapely
                    else:
                        logger.warning(f"{item.tag} {item.puic} doest not have valid TrackFragments railRef {rail_info.ref} cant create geometry")

    def _construct_demarcation_markers(self):
        for item in self.get_all():
            if item.demarcation_markers is not None:
                for rail_info in item.demarcation_markers.rail_infos:
                    if hasattr(rail_info, "at_measure"):
                        rail_con = self.tree.find(rail_info.ref)
                        if rail_con is None:
                            logger.warning(f"Demarcation marker of object {item.puic} missing RailConneciton {rail_info.ref}")
                            continue
                        projection = rail_con.geometry.shapely.interpolate(rail_info.at_measure)
                        rail_info.geometry = projection
                    else:
                        logger.warning(f"{item.tag} {item.puic} doest not have valid DemarcationMarker cant create geometry")

    def _link_reffed_objects(self) -> None:
        def _check_and_warn(_imx_object, _attrib, _puic, _found_object):
            if _found_object is None:
                logger.warning(f"{_imx_object.tag} {_imx_object.puic}: {_attrib} {_puic} not present in dataset.")

        def has_puic_ancestor(_element, _root, _parent_map):
            if "puic" in _element.attrib and _element is not _root:
                return True
            parent = _parent_map.get(_element)
            while parent is not None and parent is not _root:
                if "puic" in parent.attrib:
                    return True
                parent = _parent_map.get(parent)
            return False

        def _append_relation(_puic, _imx_object, _attrib):
            found_object = self.tree.find(_puic)
            _check_and_warn(_imx_object, _attrib, _puic, found_object)
            _imx_object.reffed_objects.objects.append(ImxRelation(_attrib, _puic, found_object))

        for imx_object in self._tree.objects():
            parent_map = {c: p for p in imx_object.element.iter() for c in p}
            elements = [elem for elem in imx_object.element.iter() if not has_puic_ancestor(elem, imx_object.element, parent_map)]

            for element in elements:
                for attrib in element.attrib:
                    if attrib[-3:] == "Ref":
                        puic = element.attrib[attrib]
                        _append_relation(puic, imx_object, attrib)

                    if attrib[-4:] == "Refs":
                        for puic in element.attrib[attrib].split(" "):
                            _append_relation(puic, imx_object, attrib)

                for node in element:
                    if node.tag.endswith("Ref"):
                        attrib = ET.QName(node.tag).localname
                        puic = node.text
                        _append_relation(puic, imx_object, attrib)

                    if node.tag.endswith("Refs"):
                        if node.text is not None:
                            for puic in node.text.split(" "):
                                attrib = ET.QName(node.tag).localname
                                _append_relation(puic, imx_object, attrib)

            count_of_invalid_relations = 0

            for imx_ref_relation in imx_object.reffed_objects.objects:
                if imx_ref_relation.reffed_object is None:
                    count_of_invalid_relations += 1

            if count_of_invalid_relations != 0:
                imx_object.reffed_objects.has_invalid_relations = True

    def _link_rail_connection_info(self):
        for imx_object in self._tree.objects():
            if imx_object.rail_connection_infos:
                for item in imx_object.rail_connection_infos:
                    found_object = self.tree.find(item.ref)
                    if found_object:
                        item.present_in_dataset = True

                        # if isinstance(item, RailConnectionInfoPoint):
                        #     item._projection = MeasureLineString(found_object.shapely.coords).\
                        #         project(imx_object.shapely, imx_object.geometry.azimuth)
                        #
                        # elif isinstance(item, RailConnectionInfoLine):
                        #     rail_con = MeasureLineString(found_object.shapely.coords)
                        #     if item.from_measure < item.to_measure:
                        #         item._projection = rail_con.cut_profile(item.from_measure, item.to_measure)
                        #     else:
                        #         item._projection = rail_con.cut_profile(item.to_measure, item.from_measure)

                    else:
                        item.present_in_dataset = False
                        logger.warning(f"RailConnection with puic {item.ref} not present in dataset.")
            # print()

    def _register_children(self):
        for imx_object in self._tree.objects():
            children = imx_object.element.findall(".//*[@puic]")
            if children is not None:
                imx_object.children = [self._tree.find(child.get("puic")) for child in children]

    def get_all(self, areas: Optional[List[AreaNameEnum]] = None) -> List[ImxObject]:
        """
        Get all value objects in repository.

        Returns:
            (List[ImxObject]): List of all value objects.
        """
        all_objects = list(self._tree.objects())
        if areas is None:
            return all_objects
        else:
            return [item for item in all_objects if item.area is None or item.area.name in areas]

    def get_by_puic(self, puic: str) -> Optional[ImxObject]:
        """
        Get value objects by PUIC.

        Returns:
            (ImxObject): Value object.
        """
        return self._tree.find(puic)

    def get_all_types(self) -> List[str]:
        """
        Get all value object types present in repository.

        Returns:
            (List[str]): List of value objects types.
        """
        return list(set([item.tag for item in self.get_all()]))

    def get_by_types(self, object_types: List[str]) -> List[ImxObject]:
        """
        Get all value object of a specific type.

        Returns:
            (List[ImxObject]): List of all value objects.
        """
        return [item[0] for item in self.tree._tree_dict.values() if item[0].tag in object_types]

    def get_all_paths(self):
        """
        Get all value object paths present in repository.

        Returns:
            (List[str]): List of value objects paths.
        """
        return list(set([item.path for item in self.get_all()]))

    def get_by_paths(self, object_paths: List[str]) -> List[ImxObject]:
        """
        Get all value object of a specific path.

        Returns:
            (List[ImxObject]): List of all value objects.
        """
        return [item[0] for item in self.tree._tree_dict.values() if item[0].path in object_paths]

    def get_pandas_df(self, object_type_or_path: Optional[str] = None, puic_as_index=True) -> pd.DataFrame:
        """
        Get Pandas dataframe of one value object type or limited view of all objects.

        When using a object type or path all properties will be flatted, when getting a dataframe of all object most attributes will be stripped
            except some metadata. In both cases it will include parent puic, path and classified area.

        Args:
            object_type_or_path: path or imx type to get df off
            puic_as_index: if true puic value will lbe the index

        Returns:
            (pd.DataFrame): pandas dataframe of the object properties
        """
        if object_type_or_path is None:
            props_in_overview = [
                "Location.GeographicLocation.@accuracy",
                "Location.GeographicLocation.@dataAcquisitionMethod",
                "Metadata.@isInService",
                "Metadata.@lifeCycleStatus",
                "Metadata.@source",
            ]
            records = []
            for item in self.get_all():
                props = {key: value for key, value in item.properties.items() if key in props_in_overview}
                records.append(
                    {
                        "puic": item.puic,
                        "area": item.area.name if item.area is not None else "No Area classified",
                        "path": item.path,
                        "parent": item.parent.puic if item.parent is not None else "",
                        "name": item.name,
                    }
                    | props
                )

        else:
            if "." in object_type_or_path:
                value_objects = self.get_by_paths([object_type_or_path])
            else:
                value_objects = self.get_by_types([object_type_or_path])

            records = []
            for item in value_objects:
                records.append(
                    {
                        "puic": item.puic,
                        "area": item.area.name if item.area is not None else "No Area classified",
                        "path": item.path,
                        "parent": item.parent.puic if item.parent is not None else "",
                        "name": item.name,
                    }
                    | item.properties
                )

        df = pd.DataFrame.from_records(records)
        if len(records) != 0 and puic_as_index:
            df["index"] = df["puic"]
            df.set_index("index", inplace=True)
            df = df.fillna("")
        return df

    def get_pandas_df_dict(self, key_based_on_type: bool = False) -> Dict[str, pd.DataFrame]:
        """
        Get a dictionary of Pandas dataframe of all type bases on keys or path.

        Args:
            key_based_on_type: if true key based on type, else on path

        Returns: dictionary of pandas dataframe

        """
        out_dict = {}
        if key_based_on_type:
            for imx_type in self.get_all_types():
                out_dict[imx_type] = self.get_pandas_df(imx_type)

        for imx_path in self.get_all_paths():
            out_dict[imx_path] = self.get_pandas_df(imx_path)

        return out_dict

    def get_geojson(self, object_type_or_path: str, in_rd: bool = False) -> GeoJsonFeatureCollection:
        """
        Get a geojson feature collection of certain type or path.

        Args:
            object_type_or_path: object type or path
            in_rd: if True returns coordinates in RD else WGS84

        Return:
             GeojsonFeatureCollection

        """
        if "." in object_type_or_path:
            records = self.get_by_paths([object_type_or_path])
        else:
            records = self.get_by_types([object_type_or_path])

        if in_rd:
            features = [GeoJsonFeature(properties=record.properties, geometry_list=[record.shapely]) for record in records]
        else:
            features = [
                GeoJsonFeature(properties=record.properties, geometry_list=[ShapelyTransform.rd_to_wgs(record.shapely)]) for record in records
            ]

        feature_collection = GeoJsonFeatureCollection(features)
        return feature_collection

    def get_geojson_dict(self, key_based_on_type: bool = False, in_rd: bool = False) -> Dict[str, GeoJsonFeatureCollection]:
        """
        Get a dictionary of geojson feature collection of all types or paths.

        Args:
            key_based_on_type:
            in_rd: if True returns coordinates in RD else WGS84

        Returns:
             Dict of geojson feature collections.
        """
        out_dict = {}
        if key_based_on_type:
            for imx_type in self.get_all_types():
                out_dict[imx_type] = self.get_geojson(imx_type, in_rd)

        for imx_path in self.get_all_paths():
            out_dict[imx_path] = self.get_geojson(imx_path, in_rd)

        return out_dict

    def _create_section_geometry(self):
        section_querier = SectionGeometryGraphQuery(self.graph)
        for item in self._tree.objects():
            if item.tag == "AxleCounterSection":
                item.geometry.shapely = section_querier.get_section(item, ["AxleCounterDetectionPointRefs"]).geometry
            elif item.tag == "TrackCircuit":
                item.geometry.shapely = section_querier.get_section(item, ["InsulatedJointRefs"]).geometry

    # def _set_azimuth_from_functional_direction(self):
    #     for item in self._tree.objects():
    #         if not item.geometry.azimuth and item.rail_connection_infos is not None:
    #             if len(item.rail_connection_infos.rail_infos) == 1:
    #
    #                 if hasattr(item.rail_connection_infos.rail_infos[0], "direction"):
    #                     if item.rail_connection_infos.rail_infos[0].direction == "Upstream":
    #                         pass
    #                     if item.rail_connection_infos.rail_infos[0].direction == "Downstream":
    #                         pass
    #                     else:
    #                         item.geometry.azimuth = 999



from typing import Protocol

from imxInsights.domain.models.imxEnums import AreaNameEnum
from imxInsights.repo.tree.objectTree import ImxObject


class ImxSituationRepoMinimalProtocol(Protocol):
    # fmt: off
    def get_by_types(self, object_types: list[str]) -> list[ImxObject]: ...  # noqa E704

    def get_by_puic(self, puic: str) -> ImxObject | None: ...  # noqa E704


    def get_all(self, areas: list[AreaNameEnum] | None) -> list[ImxObject]: ...  # noqa E704
    # fmt: on



import re
from dataclasses import dataclass
from typing import Any, List, Optional, Set, Tuple, TypeVar

from imxInsights.repo.imxRepo import SituationRepo
from imxInsights.report.imxObjectDisplay import ImxObjectDisplay, ImxPuicDisplay

T = TypeVar("T")


@dataclass
class GuidComparisonResult:
    added_guids: List[str]
    removed_guids: List[str]
    unchanged_guids: List[str]


class ImxRefDisplay:
    def __init__(self, imx_situation_init: SituationRepo, imx_situation_new: SituationRepo) -> None:
        self._puic_display = ImxObjectDisplay(imx_situation_init, imx_situation_new)

    def get_display(self, value: T) -> T:
        if not isinstance(value, str):
            return value

        status, guids = self.get_uuid_from_diff_str(value)

        if not guids:
            return value

        guid_comparison_result = self._process_guids(value, guids)

        removed = self._nice_display_comprehension(guid_comparison_result.removed_guids)
        unchanged = self._nice_display_comprehension(guid_comparison_result.unchanged_guids)
        added = self._nice_display_comprehension(guid_comparison_result.added_guids)

        return self._format_ref_output(removed, added, unchanged)

    def _nice_display_comprehension(self, guids: List[str]) -> list[ImxPuicDisplay]:
        return [self._puic_display.get_nice_display(guid) for guid in guids]

    def _process_guids(self, value: Any, guids: List[str]) -> GuidComparisonResult:
        # todo: use status from re magic instead of if else below in code

        if " -> " in value:
            left, right = value.split(" -> ")
            __, left_guids = self.get_uuid_from_diff_str(left)
            __, right_guids = self.get_uuid_from_diff_str(right)
            return GuidComparisonResult(*self._compare_guid_lists(left_guids, right_guids))

        else:
            if value.startswith("++"):
                return GuidComparisonResult(added_guids=guids, removed_guids=[], unchanged_guids=[])
            elif value.startswith("--"):
                return GuidComparisonResult(added_guids=[], removed_guids=guids, unchanged_guids=[])
            else:
                return GuidComparisonResult(added_guids=[], removed_guids=[], unchanged_guids=guids)

    @staticmethod
    def _format_ref_output(removed: List[ImxPuicDisplay], added: List[ImxPuicDisplay], unchanged: List[ImxPuicDisplay]):
        output = []

        if len(removed) == 1 and len(added) == 1 and len(unchanged) == 0:
            return f"{removed[0]} -> {added[0]}"

        if removed:
            output.append("-- " + "\n-- ".join(str(obj) for obj in removed))
        if added:
            output.append("++ " + "\n++ ".join(str(obj) for obj in added))
        if unchanged:
            output.append("\n".join(str(obj) for obj in unchanged))

        # Joining all categories with newline separators
        return "\n".join(output)

    @staticmethod
    def _compare_guid_lists(old_guids: List[str], new_guids: List[str]) -> Tuple[Set, Set, Set]:
        # todo: move to helper
        old_set = set(old_guids)
        new_set = set(new_guids)

        common = old_set.intersection(new_set)
        unique_to_old = old_set - new_set
        unique_to_new = new_set - old_set

        return unique_to_new, unique_to_old, common

    @staticmethod
    def get_uuid_from_diff_str(value: str) -> Tuple[Optional[str], List[str]]:
        guid_pattern = r"[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-4[0-9a-fA-F]{3}-[89abAB][0-9a-fA-F]{3}-[0-9a-fA-F]{12}"
        status_pattern = r"^(?P<status>\+\+|--)(?=\s)|(?<=\s)(?P<transition>->)(?=\s)"
        uuids_pattern = rf"{guid_pattern}"

        # Find all UUIDs first
        uuids = re.findall(uuids_pattern, value)

        # Check for status or transition
        status_match = re.search(status_pattern, value)
        if status_match:
            status = status_match.group("status") if status_match.group("status") else status_match.group("transition")
        else:
            status = None

        return (status, uuids)

    def get_parent_display(self, value: T) -> T:
        if not isinstance(value, str):
            return value
        return self._puic_display.get_parent_display(value)



import hashlib
import json
import os
from pathlib import Path
from typing import Dict, List, Optional

from ruamel.yaml import YAML

# commented out: code that is not used but could be handy.


def yaml_load(file_path):
    """
    Load a YAML file and return its contents as a dictionary.

    Args:
        file_path (str): The path to the YAML file.

    Returns:
        dict: The contents of the YAML file as a dictionary.
    """
    yaml = YAML()
    yaml.default_flow_style = True
    with open(file_path, "r") as stream:
        return yaml.load(stream)


def yaml_save(data, out_file_path):
    """
    Save a dictionary as a YAML file.

    Args:
        data (dict): The dictionary to be saved as YAML.
        out_file_path (str): The path to save the YAML file.
    """
    yaml = YAML()
    yaml.default_flow_style = True
    with open(out_file_path, "w") as outfile:
        yaml.dump(data, outfile)


def get_file_path(current_path: str, path: str) -> str:
    my_dir = os.path.dirname(current_path)
    config_file_path = os.path.join(my_dir, path)
    return config_file_path


# def is_float(element: Any) -> bool:
#     """
#     Check if a given value can be converted to a float.
#
#     :param element: The value to check.
#     :type element: any
#     :return: True if the value can be converted to a float, False otherwise.
#     :rtype: bool
#     """
#     if element is None:
#         return False
#     try:
#         float(element)
#         return True
#     except ValueError:
#         return False


def hash_dict_ignor_nested(dictionary: Dict):
    """
    Compute the SHA-1 hash of the dictionary's non-nested values.

    This function takes a dictionary as input and computes the SHA-1 hash of its
    content, excluding nested dictionaries. It extracts non-dictionary values
    from the input dictionary and creates a new dictionary containing only those
    values. Then, it sorts the keys of the new dictionary and computes the SHA-1
    hash of the resulting JSON-encoded string.

    Args:
        dictionary (Dict): The dictionary whose content should be hashed.

    Returns:
        str: A hexadecimal string representing the SHA-1 hash of the non-nested
             values in the dictionary.

    Examples:
        >>> input_dict = {'a': 1, 'b': {'c': 2, 'd': 3}, 'e': 4}
        >>> hash_value = hash_dict_ignor_nested(input_dict)
        >>> hash_value
        'a1e3d40cd5262ccef85a17b7698b9d5e8a61a5f6'

    Note:
        This function excludes nested dictionaries when computing the hash,
        focusing only on non-dictionary values. If the input dictionary contains
        nested dictionaries, their content will not be included in the hash.
    """
    new_dict = {}
    for key, value in dictionary.items():
        if not isinstance(value, dict):
            new_dict[key] = value

    hash_object = hashlib.sha1(json.dumps(new_dict, sort_keys=True).encode())
    return hash_object.hexdigest()


def flatten_dict(data_dict: Dict[str, Dict | str | List], skip_key: Optional[str] = "@puic", prefix="", sep=".") -> dict[str, str]:
    def _custom_sorting(key_, remaining_: List[Dict]):
        mapping = {
            "RailConnectionInfo": "@railConnectionRef",
            "Announcement": "@installationRef",
        }
        if key_ in mapping.keys():
            return sorted(remaining_, key=lambda x: x[mapping[key_]])
        else:
            return sorted(remaining_, key=hash_dict_ignor_nested)

    result: Dict[str, str] = {}

    # Skip root node if this is a recursive call and key is found
    if prefix and skip_key in data_dict:
        return result

    for key, value in data_dict.items():
        if not isinstance(value, list) and not isinstance(value, dict):
            result[f"{prefix}{key}"] = value
            continue

        new_prefix = f"{prefix}{key}{sep}"

        if isinstance(value, list) and len(value) > 0 and not isinstance(value[0], dict):
            # add index and add to current.
            for i, child in enumerate(value):
                result[f"{new_prefix}{i}"] = child
            continue

        # Multiple children -> list of dicts. Convert single dict to list with dict.
        if isinstance(value, dict):
            value = [value]

        assert len(value) > 0 and isinstance(value[0], dict)

        # Filter children with skip_key.
        remaining = list[dict]() if skip_key is not None else value
        if skip_key is not None:
            for child in value:
                if skip_key in child:
                    continue
                remaining.append(child)

        # Add index for each child if >1 and recurse, order can be changed in xml, so sort before indexing..
        if len(remaining) > 1:
            # sorting is done on a specified key.
            # if no key is present make hash of attributes, if no attributes hash first node attributes...
            remaining = _custom_sorting(key, remaining)

        for i, child in enumerate(remaining):
            child_prefix = f"{new_prefix}{i}{sep}" if len(remaining) > 1 else new_prefix
            flattened = flatten_dict(child, skip_key=skip_key, prefix=child_prefix, sep=sep)
            result = result | flattened

    return result


def hash_sha256(path: Path):
    """
    Calculate the SHA-256 hash sum of a file located at the specified path.

    This function takes a `Path` object representing the path to a file and
    calculates the SHA-256 hash sum of the file's contents. It returns the
    hash sum as a hexadecimal string.

    Args:
        path (Path): The path to the file for which the SHA-256 hash sum
            should be calculated.

    Returns:
        str: A hexadecimal string representing the SHA-256 hash sum of the file.

    Examples:
        >>> file_path = Path("example.txt")
        >>> hash_value = hash_sha256(file_path)
        >>> hash_value
        'a12b3c4d5e6f7a8b9c0d1e2f3a4b5c6d7e8f9a0b1c2d3e4f5a6b7c8d9e0f1a2b3'

    Note:
        This function reads the entire contents of the file into memory
        to calculate the hash sum. For large files, this may consume a
        significant amount of memory. Make sure to handle large files
        appropriately when using this function.

    """
    return f"{hashlib.sha256(path.read_bytes()).hexdigest()}"


def check_substring_in_keys(dictionary: dict, substring: str) -> bool:
    for key in dictionary.keys():
        if substring in key:
            return True
    return False



from datetime import datetime, timezone
from pathlib import Path
from typing import Optional

import pandas as pd
from shapely import LineString, Point, Polygon

from imxInsights.domain.models.imxEnums import ImxSituationsEnum
from imxInsights.domain.models.imxSituations import ImxProject, ImxSituation
from imxInsights.repo.imxRepo import SituationRepo
from imxInsights.report.file_info import write_situation_info
from imxInsights.utils.geodata_gis import GisRailDataLayer, get_gis_rail_data_layer
from imxInsights.utils.helpers import hash_sha256
from imxInsights.utils.log import logger
from imxInsights.utils.xml_helpers import XmlFile


class Imx:
    """
    Imx main object is used as an entry point for imx files.

    Args:
        imx_file_path (str): File path of the IMX file.

    Attributes:
        situation (Optional[ImxSituation]): The situation of an IMX file if present.
        project (Optional[ImxProject]): The ImxProject of an IMX file if present.
    """

    def __init__(self, imx_file_path: str, add_graph: bool = True, gis_service: GisRailDataLayer | bool = False):
        # todo: make add graph geometry as bool
        logger.info(f"Loading file {imx_file_path}")

        self.file_path = Path(imx_file_path)
        self.file_hash = hash_sha256(self.file_path)
        self._xml_file = XmlFile(self.file_path)
        self._imx_version: str = self._xml_file.root.find("[@imxVersion]").attrib["imxVersion"]
        logger.info(f"IMX version: {self.imx_version}")

        self.situation: Optional[ImxSituation] = None
        self.project: Optional[ImxProject] = None
        if self._xml_file.root.find(".//{*}Project") is not None:
            self.project = ImxProject(self._xml_file, add_graph)
        else:
            self.situation = ImxSituation(self._xml_file, add_graph)

        logger.info("IMX parsed, DONE!")

        if isinstance(gis_service, GisRailDataLayer):
            self._add_km(gis_service)
        elif gis_service is True:
            gis_service = get_gis_rail_data_layer()
            self._add_km(gis_service)

    def _add_km(self, gis_layer: GisRailDataLayer):
        logger.info("add km value to objects in all situations")

        # todo: make async
        if self.project is not None:
            if self.project.initial_situation is not None:
                for item in self.project.initial_situation.tree.objects():
                    self._add_km_to_all_items(item, gis_layer)
            if self.project.new_situation is not None:
                for item in self.project.new_situation.tree.objects():
                    self._add_km_to_all_items(item, gis_layer)
        elif self.situation is not None:
            for item in self.situation.situation.tree.objects():
                self._add_km_to_all_items(item, gis_layer)
        logger.info("done adding km values")

    def _add_km_to_all_items(self, item, gis_layer):
        item.km_values = []
        if isinstance(item.shapely, Point):
            item.km_values.append(gis_layer.get_km_linear_referencing(item.shapely.x, item.shapely.y))
        elif isinstance(item.shapely, LineString):
            item.km_values.append(gis_layer.get_km_linear_referencing(item.shapely.coords[0][0], item.shapely.coords[0][1]))
            item.km_values.append(gis_layer.get_km_linear_referencing(item.shapely.coords[-1][0], item.shapely.coords[-1][1]))
        elif isinstance(item.shapely, Polygon):
            if not item.shapely.is_empty:
                item.km_values.append(gis_layer.get_km_linear_referencing(item.shapely.centroid.x, item.shapely.centroid.y))

    @property
    def imx_version(self) -> str:
        """Returns the IMX version of the file, *read only property*."""
        return self._imx_version

    def get_situation_repository(self, situation_type: ImxSituationsEnum) -> Optional[SituationRepo]:
        """
        Get a specific situation repository  from an IMX project or situation file.

        Args:
            situation_type (ImxSituationsEnum): The IMX situation to get the repo.

        Returns:
            (SituationRepo): The given situation if exists in file.

        """
        if situation_type == ImxSituationsEnum.InitialSituation:
            if self.project is not None:
                return self.project.initial_situation
        elif situation_type == ImxSituationsEnum.NewSituation:
            if self.project is not None:
                return self.project.new_situation
        else:
            if self.situation is not None:
                return self.situation.situation
        return None

    @staticmethod
    def _reshape_df_worksheet(df, worksheet, filter=True):
        (max_row, max_col) = df.shape
        worksheet.set_column(0, max_col - 1, 12)
        if filter:
            worksheet.autofilter(0, 0, max_row, max_col)
        worksheet.autofit()
        worksheet.freeze_panes(1, 0)

    def generate_population_excel(self, file_path: str, situation: ImxSituationsEnum):
        repo = self.get_situation_repository(situation)

        writer = pd.ExcelWriter(file_path, engine="xlsxwriter")
        workbook = writer.book

        worksheet_info = workbook.add_worksheet("info")
        worksheet_info.set_column(0, 0, 25)
        worksheet_info.set_column(1, 1, 150)

        worksheet_info.write(1, 0, "process datestamp")
        worksheet_info.write(1, 1, datetime.now().astimezone(timezone.utc).strftime("%Y-%m-%d %H:%M:%S.%f"))

        write_situation_info(worksheet_info, repo, 2, "a")

        overview_records = []
        overview_keys_to_keep = [
            "puic",
            "path",
            "area",
            "name",
            "parent",
            "Metadata.@originType",
            "Metadata.@source",
            "Metadata.@registrationTime",
            "Metadata.@lifeCycleStatus",
            "Metadata.@isInService",
            "Location.GeographicLocation.@dataAcquisitionMethod",
            "Location.GeographicLocation.@accuracy",
        ]

        df_dict = repo.get_pandas_df_dict()

        pivot_index = [
            "path",
            "Metadata.@lifeCycleStatus",
            "Location.GeographicLocation.@accuracy",
            "Location.GeographicLocation.@dataAcquisitionMethod",
        ]

        for key, df in df_dict.items():
            records_ = df.to_dict(orient="records")

            for record_ in records_:
                records_to_add = {key: value for key, value in record_.items() if key in overview_keys_to_keep} | {"counter": 1}
                if records_to_add["parent"] == "":
                    records_to_add["_parent"] = records_to_add["puic"]
                else:
                    records_to_add["_parent"] = records_to_add["parent"]

                for item in pivot_index:
                    if item not in records_to_add.keys():
                        records_to_add[item] = ""

                overview_records.append(records_to_add)

        df = pd.DataFrame.from_records(overview_records)

        # create pivot
        table = pd.pivot_table(
            df,
            values="counter",
            index=pivot_index,
            columns=["area"],
            aggfunc="count",
            fill_value=" ",
        )
        table.to_excel(writer, sheet_name="population_pivot", index=True)
        worksheet = writer.sheets["population_pivot"]
        self._reshape_df_worksheet(table, worksheet, filter=False)

        temp_df = df["path"].str.split(".", expand=True)
        df["main_type"] = temp_df[0]
        del df["counter"]

        df_grouped = df.groupby(["main_type", "_parent", "path"], group_keys=True).apply(lambda group: group)
        df_grouped = df_grouped.assign(row_number=range(len(df)))
        df = df_grouped.set_index("row_number")

        first_column = df.pop("main_type")
        df.insert(2, "main_type", first_column)
        del df["_parent"]

        df.to_excel(writer, sheet_name="all_objects", startrow=0, startcol=0, index=1)
        worksheet = writer.sheets["all_objects"]
        self._reshape_df_worksheet(df, worksheet)

        for key, df in sorted(df_dict.items()):
            records_ = df.to_dict(orient="records")
            for record_ in records_:
                records_to_add = {key: value for key, value in record_.items() if key in overview_keys_to_keep}
                overview_records.append(records_to_add)

            sheet_name = f"{key[:14]}...{key[-14:]}" if len(key) > 30 else key
            df.to_excel(writer, sheet_name=sheet_name)

            worksheet = writer.sheets[sheet_name]
            self._reshape_df_worksheet(df, worksheet)

        worksheet_info.activate()

        writer.close()



from __future__ import annotations

from typing import List, Optional, Tuple

from lxml import etree as ET
from shapely.geometry.base import BaseGeometry

from imxInsights.domain.area.area import Area
from imxInsights.domain.area.areaClassifiedTypeEnum import AreaClassifiedTypeEnum
from imxInsights.domain.area.imxProjectArea import ImxProjectArea
from imxInsights.repo.imxRepo import ImxObject
from imxInsights.utils.log import logger
from imxInsights.utils.shapely_helpers import (
    gml_linestring_to_shapely,
    gml_point_to_shapely,
    gml_polygon_to_shapely,
)
from imxInsights.utils.xml_helpers import find_coordinates, find_puic, trim_tag


class AreaClassifier:
    """
    ImxProjectArea classifier to determinate the widest area that a object is in.

    Args:
        areas (List[Area]): list of all area that can be used to classify

    """

    def __init__(self, areas: List[ImxProjectArea]):
        self._areas: List[Area] = [Area(area) for area in areas]
        if len(self._areas) == 0:
            raise ValueError("No user/work/context or other area passed to classifier!")  # pragma: no cover

    def determine_area(self, geometry: BaseGeometry, element: Optional[ET.Element] = None) -> Area:
        for area in self._areas:
            if area.intersects(geometry):
                if area.name == "NoAreaMatch":
                    # todo: fix that it's not returning unknown all the time... not sure why it does.
                    outside = find_puic(element) if element is not None else "<unknown>"
                    logger.warning(f"Object {outside} lies outside context area, marked as NoAreaMatch")
                return area

    def categorize_element(self, imx_object: ImxObject) -> Tuple[Optional[Area], Optional[AreaClassifiedTypeEnum]]:
        """
        Determines which of the areas intersect with this element.

        If multiple coordinates are found inside this entity which do not belong to any other entity (e.g. mathematical point on switch),
        all areas are computed and the widest is chosen.

        Classify strategie:
         - use gml geometry or constructed railConnection geometry
         - use track fragments and demarcation markers
         - use reffed objects

        Args:
            imx_object (ImxObject): imx value object to categorize

        Returns:
            Tuple[(Optional[Area],  (Optional[str]]): widest area

        Todo:
         - move logic to set geometry to imx object, including AreaClassifiedTypeEnum as a generic enum

        """
        areas = list[Area]()
        classify_type: Optional[AreaClassifiedTypeEnum] = None
        coordinates = find_coordinates(imx_object.element)
        if imx_object.element.tag == "{http://www.prorail.nl/IMSpoor}RailConnection":
            classify_type = AreaClassifiedTypeEnum.CONSTRUCTED_GEOMETRY
            areas.append(self.determine_area(geometry=imx_object.shapely))

        elif len(coordinates) > 0:
            classify_type = AreaClassifiedTypeEnum.OBJECT_GEOMETRY
            for coords in coordinates:
                geometry = self._parse_coordinates(coords)
                areas.append(self.determine_area(geometry=geometry, element=coords))

        else:
            if imx_object.track_fragments is not None:
                classify_type = AreaClassifiedTypeEnum.TRACK_FRAGMENTS_AND_DEMARCATION_MARKERS
                for track_fragment in imx_object.track_fragments.rail_infos:
                    areas.append(self.determine_area(geometry=track_fragment.geometry))

            if imx_object.demarcation_markers is not None:
                classify_type = AreaClassifiedTypeEnum.TRACK_FRAGMENTS_AND_DEMARCATION_MARKERS
                for demarcation_marker in imx_object.demarcation_markers.rail_infos:
                    areas.append(self.determine_area(geometry=demarcation_marker.geometry))

        if len(areas) == 0:
            if imx_object.reffed_objects.has_invalid_relations:
                classify_type = AreaClassifiedTypeEnum.FAILED
                logger.warning(f"{imx_object.tag} {imx_object.puic} ")
                return None, classify_type

            if len(imx_object.reffed_objects.objects) > 0:
                classify_type = AreaClassifiedTypeEnum.RELATED_OBJECTS
                for ref_relation in imx_object.reffed_objects.objects:
                    if ref_relation.present_in_dataset and not ref_relation.reffed_object.shapely.is_empty:
                        areas.append(self.determine_area(geometry=ref_relation.reffed_object.shapely))

        if len(areas) == 0:
            logger.warning(f"{imx_object.tag} {imx_object.puic} can't classify area.")
            return None, classify_type

        elif len(areas) == 1 or len(set(areas)) == 1:
            return areas[0], classify_type

        widest_area = self._areas[max(map(self._areas.index, areas))]
        return widest_area, classify_type

    @staticmethod
    def _parse_coordinates(coordinates: ET.Element) -> BaseGeometry:
        assert coordinates.tag.endswith("coordinates"), "xml element does not have <gml:coordinates/>"
        parent_tag = trim_tag(coordinates.getparent().tag)
        match parent_tag:
            case "Point":
                return gml_point_to_shapely(coordinates.text)
            case "LineString":
                return gml_linestring_to_shapely(coordinates.text)
            case "LinearRing":
                return gml_polygon_to_shapely(coordinates.text)
            case _:  # pragma: no cover
                raise ValueError("Cannot parse coordinates for element")



from enum import Enum


class AreaClassifiedTypeEnum(str, Enum):
    """
    Possible classification types.

    Attributes:
        OBJECT_GEOMETRY (str): ....
        CONSTRUCTED_GEOMETRY (str): ....
        TRACK_FRAGMENTS_AND_DEMARCATION_MARKERS (str): ....
        RELATED_OBJECTS (str): ....
        FAILED (str): ....
    """

    OBJECT_GEOMETRY = "objects gml:geometry is used"
    CONSTRUCTED_GEOMETRY = "using constructed geometry."
    TRACK_FRAGMENTS_AND_DEMARCATION_MARKERS = "track fragments and demarcation markers"
    RELATED_OBJECTS = "related objects geometry"
    FAILED = "common strategy failed because of reasons"



# import networkx as nx
#
# from imxInsights.graph.queries.sectionGeometryQuery import SectionGeometryGraphQuery
#
# # class GraphQueries:
# #     def __init__(self, g: nx.DiGraph):
# #         self.SectionGeometryGraphQuery: SectionGeometryGraphQuery = SectionGeometryGraphQuery(g)



from __future__ import annotations

import warnings

# from pathlib import Path
from typing import Iterable, List, Optional

from lxml import etree as ET
from shapely import GeometryCollection
from shapely.geometry.base import BaseGeometry

from imxInsights.domain.area.areaClassifiedTypeEnum import AreaClassifiedTypeEnum
from imxInsights.domain.area.imxProjectArea import ImxProjectArea
from imxInsights.domain.models.imxGeometry import ImxGeometry
from imxInsights.domain.models.imxMetadata import ImxObjectMetadata
from imxInsights.domain.models.imxPuic import ImxPuic, PuicConstructedByAttribute
from imxInsights.domain.models.ImxRailConnectionInfo import RailConnectionInfos
from imxInsights.domain.models.imxRelation import ImxReffedObjects
from imxInsights.repo.tree.imxTopoTreeObject import ImxTopoObject
from imxInsights.utils.helpers import flatten_dict, get_file_path, yaml_load
from imxInsights.utils.xml_helpers import (
    find_parent_entity,
    lxml_element_to_dict,
    trim_tag,
)


class ImxObject:
    """
    Imx value object.

    Every imx object of interest is parsed as a value object.

    Args:
        element (ET.Element):
        parent (Optional[ET.Element]):
        constructed_puic (Optional[PuicConstructedByAttribute]):

    Attributes:
        metadata (ImxObjectMetadata): contains value object metadata.
        geometry (ImxGeometry): contains geometry of value object

        rail_connection_infos (Optional[RailConnectionInfos]): contains all railConnectionInfo objects in value object.
        track_fragments (Optional[RailConnectionInfos]): contains all railConnectionInfo objects of the track fragments in value object.
        demarcation_markers (Optional[RailConnectionInfos]): contains all railConnectionInfo objects of the demarcation markers in value object.

        reffed_objects (ImxReffedObjects): reffed objects wrapper.
        children (List[ImxObject]): child value objects.
        topo_object (ImxTopoObject): corresponding topo object.
        area (Optional[ImxProjectArea]): classified area set by AreaClassifier.
        area_classify_type (AreaClassifiedTypeEnum): methode used to classify area.

    """

    def __init__(
        self,
        element: ET.Element,
        parent: Optional[ET.Element] = None,
        constructed_puic: Optional[PuicConstructedByAttribute] = None,
    ):
        self._element: ET.Element = element

        self.parent: Optional[ImxObject] = parent
        if constructed_puic is not None:
            self._puic: ImxPuic = ImxPuic.from_element(self._element, constructed_puic)
        else:
            self._puic = ImxPuic.from_element(self._element)

        self.properties: dict[str, str] = flatten_dict(lxml_element_to_dict(self._element))
        self.metadata = ImxObjectMetadata(self.element)
        self.geometry: Optional[ImxGeometry] = None
        self.rail_connection_infos: Optional[RailConnectionInfos] = RailConnectionInfos.from_element(self.element, "RailConnectionInfo")

        self.track_fragments: Optional[RailConnectionInfos] = RailConnectionInfos.from_element(self.element, "TrackFragment")
        self.demarcation_markers: Optional[RailConnectionInfos] = RailConnectionInfos.from_element(self.element, "DemarcationMarker")

        self.children = List[ImxObject]
        self.reffed_objects: ImxReffedObjects = ImxReffedObjects()
        self.topo_object: Optional[ImxTopoObject] = None

        self.area: Optional[ImxProjectArea] = None
        self.area_classify_type: Optional[AreaClassifiedTypeEnum] = None
        self.km_values: Optional[List[str]] = None

    def __repr__(self) -> str:
        return f"<ImxObject {self.path} puic={self.puic} name='{self.name}'/>"

    @property
    def element(self) -> ET.Element:
        return self._element

    @property
    def tag(self) -> str:
        """Returns the ImxObject type of the value objectm (same as xml tag), *read only property*."""
        return trim_tag(self._element.tag)

    @property
    def path(self) -> str:
        """Returns the Object path of the value object, *read only property*."""
        tags = [parent.tag for parent in self._parents_generator()]
        tags.reverse()
        tags.append(self.tag)
        return ".".join(tags)

    @property
    def puic(self) -> str:
        """Returns the Puic or constructed puic of the value object, *read only property*."""
        return str(self._puic)

    @property
    def name(self) -> str:
        """Returns the name of the value obejct if present else empty string, *read only property*."""
        return str(self._element.get("name")) if "name" in self._element.keys() else ""

    @property
    def shapely(self) -> BaseGeometry:
        """Returns the Shapely object that represent the main geometry of the value object, *read only property*."""
        if self.geometry is not None:
            return self.geometry.shapely
        return GeometryCollection()

    def _parents_generator(self) -> Iterable[ImxObject]:
        parent = self.parent
        while True:
            if parent is None:
                break
            yield parent
            parent = parent.parent

    def register_topo_object(self, topo_object: ImxTopoObject):
        """
        Guard and register topology objects on imx object.

        Warning:
            ImxInsights repo is responsible for registering topo objects, this methode should not be used manually.

        Args:
            topo_object (ImxTopoObject):

        Raises:
            ValueError: If puic and topo key not same.

        """
        if topo_object.key == self.puic:
            self.topo_object = topo_object
        else:
            raise ValueError("topo object should have same key as value object")

    def can_compare(self, other: ImxObject | None) -> bool:
        """
        Checks if object can be compared whit other object, it should if it has the same puic and same object type.

        Args:
            other (ImxObject):

        Returns:
            (bool): True if can compair else False.
        """
        if other is None:
            return True

        if other.puic != self.puic:
            return False

        if other.path != self.path:
            warnings.warn(f"Cant to compare {self.path} with {other.path}, tags do not match")
            return False

        return True

    @staticmethod
    def lookup_tree_from_root_element(start: ET.Element) -> list[ImxObject]:
        """
        Factory method to initialize ImxObject lookup tree from xml root element.

        Methode generates a tree of objects given a start element by finding all objects with a puic attribute, or those specified by custom key.

        Args:
            start (ET.Element):

        """
        imx_key = "@puic"
        lookup = dict[ET.Element, ImxObject]()
        entities: list[ET.Element] = start.findall(f".//*[{imx_key}]")

        version: str = start.getroottree().getroot().attrib["imxVersion"]

        # config_file_path = f"{Path(__file__).parent.parent.parent}\\custom_puic_config.yaml"
        config_file_path = get_file_path(__file__, "../../custom_puic_config.yaml")

        custom_puic_config = yaml_load(config_file_path)
        if version == "1.2.4":
            custom_puic_config = custom_puic_config["v124"]
        elif version == "5.0.0":
            custom_puic_config = custom_puic_config["v500"]

        for entity in entities:
            parent = find_parent_entity(entity)
            if parent is not None:
                assert parent in lookup, "Expected parent in lookup"
                parent = lookup[parent]
            assert entity not in lookup
            lookup[entity] = ImxObject(parent=parent, element=entity)

        custom_attribute_puics: List[PuicConstructedByAttribute] = [
            PuicConstructedByAttribute(object_type, values[0], values[1]) for object_type, values in custom_puic_config.items()
        ]

        for custom_puic in custom_attribute_puics:
            objects_with_special_key: list[ET.Element] = start.findall(".//{*}" + custom_puic.object_type)
            for entity in objects_with_special_key:
                assert entity not in lookup
                lookup[entity] = ImxObject(parent=None, element=entity, constructed_puic=custom_puic)

        return list(lookup.values())



import asyncio
from dataclasses import dataclass
from typing import List, TypeVar, cast

import nest_asyncio
from loguru import logger
from shapely import LineString
from shapely.ops import linemerge

from imxInsights.graph.queries.imxGraphProtocol import ImxGraphProtocol
from imxInsights.repo.tree.objectTreeProtocol import ImxObjectProtocol
from imxInsights.utils.shapely_geojson import (
    GeoJsonFeature,
    GeoJsonFeatureCollection,
    dump,
)
from imxInsights.utils.shapely_helpers import ShapelyTransform

T = TypeVar("T")

nest_asyncio.apply()


@dataclass
class ImxGraphSectionQueryResult:
    section: ImxObjectProtocol
    geometry: LineString


class SectionGeometryGraphQuery:
    def __init__(self, imx_graph: ImxGraphProtocol):
        self.g: ImxGraphProtocol = imx_graph

    async def _get_section_geometry(self, refs: List[str]) -> LineString:
        refs_mapping = {uuid: [other_uuid for other_uuid in refs if other_uuid != uuid] for uuid in refs}
        line_strings = []

        # todo: make task for every mapping item so we can process them async
        for key, values in refs_mapping.items():
            from_obj = self.g.imx_situation.get_by_puic(key)
            if from_obj is None:
                logger.warning(f"Object {key} not found in situation")
                continue

            # todo: make task for every to item so we can process them async
            for item in values:
                to_obj = self.g.imx_situation.get_by_puic(item)
                if to_obj is None:
                    logger.warning(f"Object {item} not found in situation")
                    continue

                paths = self.g.get_paths_between_imx_objects(from_obj, to_obj)
                for path in paths:
                    line_strings.append(path.geometry)

        return linemerge(line_strings).buffer(0.01)

    async def _get_section_async(self, section: T, ref_type: List[str]) -> ImxGraphSectionQueryResult:
        refs = [_.key for _ in section.reffed_objects.objects if _.type in ref_type]
        line_string = await self._get_section_geometry(refs)
        return ImxGraphSectionQueryResult(section, line_string)

    async def _get_all_sections_async(self, section_type: str, ref_type: List[str]) -> List[ImxGraphSectionQueryResult]:
        sections = self.g.imx_situation.get_by_types([section_type])
        tasks = [self._get_section_async(section, ref_type) for section in sections]
        result = await asyncio.gather(*tasks)
        return cast(List[ImxGraphSectionQueryResult], result)

    def get_section(self, section: T, ref_type: List[str]) -> ImxGraphSectionQueryResult:
        return asyncio.run(self._get_section_async(section, ref_type))

    def get_all_section(self, section_type: str, ref_type: List[str]) -> List[ImxGraphSectionQueryResult]:
        return asyncio.run(self._get_all_sections_async(section_type, ref_type))

    @staticmethod
    def _create_geojson_features(query_results: List[ImxGraphSectionQueryResult]):
        features = []
        for result in query_results:
            features.append(GeoJsonFeature([ShapelyTransform.rd_to_wgs(result.geometry)], result.section.properties | {"type": result.section.tag}))
        return features

    @staticmethod
    def _save_as_geojson(fc, file_name):
        with open(file_name, "w") as file:
            dump(fc, file)

    def create_geojson_files(self):
        try:
            features = self.get_all_section("AxleCounterSection", ["AxleCounterDetectionPointRefs"])
            tester = self._create_geojson_features(features)
            self._save_as_geojson(GeoJsonFeatureCollection(tester), "AxleCounterSection.geojson")

            features = self.get_all_section("TrackCircuit", ["InsulatedJointRefs"])
            tester = self._create_geojson_features(features)
            self._save_as_geojson(GeoJsonFeatureCollection(tester), "TrackCircuit.geojson")

        except RuntimeError as e:
            print(f"Error running async function: {e}")



from dataclasses import dataclass
from enum import Enum
from typing import List, Optional

from imxInsights.repo.imxRepo import SituationRepo
from imxInsights.repo.tree.imxTreeObject import ImxObject

# todo: make tests


class InSituationEnum(Enum):
    INIT_AND_NEW = "In init and new situation"
    ONLY_INIT = "Only in init situation"
    ONLY_NEW = "Only in new situation"
    NOT_PRESENT = "Not in init and not in new situation"


@dataclass
class ImxPuicDisplay:
    path: str
    puic: str
    status: InSituationEnum
    name: str

    def display(self) -> str:
        return f"{self.path} {self.name} ({self.puic}:{self.status.name})"

    def __repr__(self) -> str:
        return self.display()

    def __str__(self) -> str:
        return self.display()


class ImxObjectDisplay:
    def __init__(self, imx_situation_init: SituationRepo, imx_situation_new: SituationRepo) -> None:
        self.init_situation = imx_situation_init
        self.new_situation = imx_situation_new

    def get_nice_display(self, puic: str) -> ImxPuicDisplay:
        init_obj, new_obj = self.init_situation.get_by_puic(puic), self.new_situation.get_by_puic(puic)

        if init_obj is None and new_obj is None:
            return ImxPuicDisplay(path="", puic=puic, status=InSituationEnum.NOT_PRESENT, name="")
        elif init_obj is None:
            return ImxPuicDisplay(path=new_obj.path, puic=new_obj.puic, status=InSituationEnum.ONLY_NEW, name=new_obj.name)
        elif new_obj is None:
            return ImxPuicDisplay(path=init_obj.path, puic=init_obj.puic, status=InSituationEnum.ONLY_INIT, name=init_obj.name)

        return ImxPuicDisplay(
            path=(new_obj or init_obj).path, puic=puic, status=self._get_status(init_obj, new_obj), name=self._get_name(init_obj, new_obj)
        )

    @staticmethod
    def _get_status(init_obj: Optional[ImxObject], new_obj: Optional[ImxObject]) -> InSituationEnum:
        if init_obj and new_obj:
            return InSituationEnum.INIT_AND_NEW
        if not init_obj and new_obj:
            return InSituationEnum.ONLY_NEW
        if init_obj and not new_obj:
            return InSituationEnum.ONLY_INIT
        return InSituationEnum.NOT_PRESENT

    @staticmethod
    def _get_name(init_obj: Optional[ImxObject], new_obj: Optional[ImxObject]) -> str:
        if not init_obj:
            return f'"{new_obj.name}"'
        if not new_obj:
            return f'"{init_obj.name}"'
        if init_obj.name == new_obj.name:
            return f'"{new_obj.name}"'
        return f'"{init_obj.name} => {new_obj.name}"' if init_obj.name else f'"{None} => {new_obj.name}"'

    @staticmethod
    def _get_all_parents(imx_object: ImxObject) -> List[ImxObject]:
        if not imx_object:
            return []
        parents = []
        while imx_object.parent:
            parents.append(imx_object.parent)
            imx_object = imx_object.parent
        return parents

    def _create_parent_path(self, obj: ImxObject) -> List[str]:
        parents = self._get_all_parents(obj) + [obj]
        return [f"<{item.name if item.name else 'NoName'} ({item.puic})/>" for item in reversed(parents)]

    def _get_formatted_path(self, obj: ImxObject) -> str:
        if not obj:
            return ""
        return ".".join(self._create_parent_path(obj)[::-1])

    def get_parent_display(self, puic_value: str) -> str:
        if not puic_value:
            return puic_value

        init_object = self.init_situation.get_by_puic(puic_value)
        init_path = self._get_formatted_path(init_object)

        new_object = self.new_situation.get_by_puic(puic_value)
        new_path = self._get_formatted_path(new_object)

        if not new_object and init_path:
            return init_path
        elif not init_object and new_object:
            return new_path
        elif init_path == new_path or init_path == []:
            return new_path

        return f"{init_path} => {new_path}"



from __future__ import annotations

from dataclasses import dataclass, field

from shapely.geometry.base import BaseGeometry

from imxInsights.domain.area.imxProjectArea import ImxProjectArea


@dataclass(frozen=True)
class Area:
    geometry: ImxProjectArea = field(repr=False, hash=False)  # referer to Area BaseGeometry does not typehint.
    name: str | None = field(kw_only=True, default=None)

    def __post_init__(self):
        if self.name is None and hasattr(self.geometry, "name"):
            super().__setattr__("name", self.geometry.name)

    def intersects(self, other: BaseGeometry) -> bool:
        return self.geometry.intersects(other)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import List, Optional

import numpy as np
from lxml import etree as ET

from imxInsights.domain.area.imxProjectArea import ImxProjectArea
from imxInsights.utils.log import logger
from imxInsights.utils.shapely_geojson import GeoJsonFeature, GeoJsonFeatureCollection
from imxInsights.utils.shapely_helpers import ShapelyTransform


class ProjectNaiadeRemarks:
    """
    Project remarks contains some Naiade configuration information.

    Info:
       - only Naiade Remarks are supported.

    """

    def __init__(self, elem: ET.Element):
        self._element = elem
        self._imx_remarks = self._get_remarks_dict()

    def _get_remarks_dict(self):
        try:
            imx_remarks = self._element.text.split(",")
            imx_remarks = dict([item.replace("'", "").split(":") for item in imx_remarks])
            imx_remarks = {x.strip().replace(" ", ""): v.strip() for x, v in imx_remarks.items()}
            return imx_remarks
        except Exception as e:  # pragma: no cover
            logger.warning(f"Metadata remarks not conform naiade format, will not be parsed. ({e})")
            return {}

    @property
    def application_version(self) -> str:
        """Returns the Naiade application version used when exporting, *read only property*."""
        if self._imx_remarks != {}:
            return str(self._imx_remarks["Applicationversion"])
        else:
            return "None"

    @property
    def environment(self) -> str:
        """Returns the Naiade environment used when exporting, *read only property*."""
        if self._imx_remarks != {}:
            return str(self._imx_remarks["Environment"])
        else:
            return "None"

    @property
    def project_type(self) -> str:
        """Returns the project type used when exporting, *read only property*."""
        if self._imx_remarks != {}:
            return str(self._imx_remarks["Projecttype"])
        else:
            return "None"

    @property
    def xsd(self) -> str:
        """Returns the xsd version used when exporting, *read only property*."""
        if self._imx_remarks != {}:
            return str(self._imx_remarks["XSD"])
        else:
            return "None"

    @property
    def validation_profile(self) -> str:
        """Returns the Naiade validation profile used when exporting, *read only property*."""
        if self._imx_remarks != {}:
            return str(self._imx_remarks["ValidationProfile"])
        else:
            return "None"


class SituationMetadata:
    """Situation metadata contains some information of the IMX data."""

    def __init__(self, elem: ET.Element):
        self._element: ET.Element = elem

    @property
    def reference_date(self) -> str:
        """
        Returns the external project reference as given in the project metadata, *read only property*.

        ***The reference date is the date that the data is valid, it could be historic or feature situation.***

        """
        return str(self._element.get("ReferenceDate"))

    @property
    def perspective_date(self) -> str:
        """
        Returns the perspective date as given in the situation metadata, *read only property*.

        *** The perspective date is the date the imx file is created and can be used to determinate if the data baseline is still up to date.***

        """
        return str(self._element.get("PerspectiveDate"))


class ProjectMetadata:
    """
    Project metadata contains some information of the IMX project.

    Attributes:
        user_area (ImxProjectArea): The imx user area.
        work_area (ImxProjectArea): The imx work area.
        context_area (ImxProjectArea): The imx context area.

    """

    def __init__(self, elem: ET.Element):
        self._element: ET.Element = elem
        self.user_area = self._areas_factory("UserArea")
        self.work_area = self._areas_factory("WorkArea")
        self.context_area = self._areas_factory("ContextArea")

    @property
    def project_areas(self) -> List[ImxProjectArea]:
        """Returns all imx project areas, *read only property*."""
        outside_area = ImxProjectArea(name="NoAreaMatch", gml_coordinates="0,0 300000,0 300000,600000 0,600000 0,0")
        return [self.user_area, self.work_area, self.context_area, outside_area]

    @property
    def external_project_reference(self) -> str:
        """Returns the external project reference as given in the project metadata, *read only property*."""
        return str(self._element.get("externalProjectReference"))

    @property
    def project_name(self) -> str:
        """Returns the project name as given in the project metadata, *read only property*."""
        return str(self._element.get("projectName"))

    @property
    def project_type(self) -> str:
        """Returns the project type as given in the project metadata, *read only property*."""
        return str(self._element.get("projectType"))

    @property
    def created_date(self) -> str:
        """
        Returns the created date as given in the project metadata, *read only property*.

        ***The created date is the date the file is mutated and saved and can be used to check if latest version.***

        """
        return str(self._element.get("createdDate"))

    @property
    def planned_delivery_date(self) -> str:
        """
        Returns the planned delivery date as given in the project metadata, *read only property*.

        ***The planned delivery date is the date the project will be live, in Dutch: indienststelling or IDS datum***

        """
        return str(self._element.get("plannedDeliveryDate"))

    @property
    def changeable_types(self) -> List[str]:
        """Returns a list of the changeable types given in the project metadata, *read only property*."""
        return str(self._element.get("ChangeableTypes")).split(",")

    def _areas_factory(self, area_name: str) -> ImxProjectArea:
        coordinates = self._element.find(f".//{{*}}{area_name}//{{*}}coordinates")
        if coordinates is None:
            raise ValueError(f"No project metadata in file {self._element.docinfo.URL}!")  # pragma: no cover

        return ImxProjectArea(name=area_name, gml_coordinates=coordinates.text)

    def area_geojson(self) -> GeoJsonFeatureCollection:
        return GeoJsonFeatureCollection(
            [
                GeoJsonFeature([ShapelyTransform.rd_to_wgs(self.user_area.polygon)], {"area": "user"}),
                GeoJsonFeature([ShapelyTransform.rd_to_wgs(self.work_area.polygon)], {"area": "work"}),
                GeoJsonFeature([ShapelyTransform.rd_to_wgs(self.context_area.polygon)], {"area": "context"}),
            ]
        )


@dataclass
class ImxObjectMetadata:
    """Stores metadata for an IMX object."""

    _element: ET.Element = field(repr=False)
    _metadata_node = False
    _location_node = False

    def __post_init__(self) -> None:
        metadata = self._element.findall("{*}Metadata")
        if len(metadata) != 0:
            self._metadata = metadata[0]
            self._metadata_node = True

        geo_location = self._element.findall("{*}Location/{*}GeographicLocation")
        if len(geo_location) != 0:
            self._location_node = True
            self._geo_location = geo_location[0]

    @staticmethod
    def _get_attribute_or_none(element: ET.Element, attribute: str) -> Optional[str]:
        if attribute in element.attrib:
            return str(element.attrib[attribute])
        else:
            return None

    @staticmethod
    def _return_string(value: str | None) -> str:
        if value:
            return str(value)
        else:
            return ""

    @property
    def origin_type(self) -> str:
        """Returns the origin type of the value object, *read only property*."""
        value = self._get_attribute_or_none(self._metadata, "originType")
        return self._return_string(value)

    @property
    def source(self) -> str:
        """Returns the source of the value object, *read only property*."""
        value = self._get_attribute_or_none(self._metadata, "source")
        return self._return_string(value)

    @property
    def life_cycle_status(self) -> str:
        """Returns the life cycle status of the value object, *read only property*."""
        value = self._get_attribute_or_none(self._metadata, "lifeCycleStatus")
        return self._return_string(value)

    @property
    def is_in_service(self) -> str:
        """Returns the is in service status of the value object, *read only property*."""
        value = self._get_attribute_or_none(self._metadata, "isInService")
        return self._return_string(value)

    @property
    def location_accuracy(self) -> float:
        """Returns the location accuracy of the value object, *read only property*."""
        accuracy = self._get_attribute_or_none(self._geo_location, "accuracy")
        if accuracy:
            return float(accuracy)
        else:
            return np.NaN

    @property
    def location_data_acquisition_method(self) -> str:
        """Returns the location data acquisition method of the value object, *read only property*."""
        value = self._get_attribute_or_none(self._geo_location, "dataAcquisitionMethod")
        return self._return_string(value)

    def has_location(self) -> bool:
        """
        Returns True if value object has location metadata.

        Returns:
            (bool):True if the object is a design object, False otherwise.

        """
        return True if self._location_node else False

    def has_metadata(self) -> bool:
        """Returns True if value object has object metadata.

        Returns:
            (bool):True if the object is a design object, False otherwise.

        """
        return True if self._metadata_node else False

    def is_design(self):
        """Returns whether the object is a design object.

        Returns:
            (bool):True if the object is a design object, False otherwise.

        """
        return True if self.is_in_service and self.location_data_acquisition_method in ["Design"] else False



from __future__ import annotations

from dataclasses import InitVar, dataclass, field

from shapely import Polygon
from shapely.geometry.base import BaseGeometry

from imxInsights.utils.shapely_helpers import gml_polygon_to_shapely


@dataclass(frozen=True)
class ImxProjectArea:
    """
    IMX project areas does have a name and coordinates, those coordinates will be parsed to a polygon.

    The touches and intersects methods use the polygon attribute of the object to check whether it touches or intersects with the Polygon
      object of the other ImxProjectArea object. The Polygon object can also be directly passed as an argument.

    !!! info "Every area has his own mutation regiem, below some context about ImxProjectAres."

        - A Imx project has a geographic scope this is the UserArea, objects within this area dont have a mutation limitation.
        - Object within the UserArea could have a interact whit a object outsidethe UserArea, those object define the WorkArea and do
          have some mutation limitations.
        - Object within the WorkArea could have relations whit object outside the WorkArea, those objects define the ContextArea and cant be changed.

    Args:
        name (str): The name of the area.
        gml_coordinates (str): The GML coordinates of the area.
        polygon (Polygon): The Shapely Polygon object representing the area.
        coordinates (InitVar[tuple[tuple[int | float, ...], ...]]): The coordinates of the area as a tuple of tuples.

    Todo:
        Use enum for project area name
    """

    name: str
    gml_coordinates: InitVar[str]
    polygon: Polygon = field(repr=False, hash=False, default=None)
    coordinates: InitVar[tuple[tuple[int | float, ...], ...]] = None

    def __post_init__(self, gml_coordinates: str, coordinates: tuple[tuple[int | float, ...], ...]):
        if self.polygon is not None:
            return  # pragma: no cover

        poly = gml_polygon_to_shapely(gml_coordinates) if gml_coordinates is not None else Polygon(coordinates) if coordinates is not None else None
        super().__setattr__("polygon", poly)

    def touches(self, area: ImxProjectArea | BaseGeometry) -> bool:
        """
        Checks if the area touches another area.

        Args:
            area (ImxProjectArea | shapley.BaseGeometry): The other area to check, could be a shapely.Polygon.

        Returns:
            (bool): True if the area touches the other area, False otherwise.
        """
        other = area.polygon if isinstance(area, ImxProjectArea) else area
        return bool(self.polygon.touches(other))

    def intersects(self, area: ImxProjectArea | BaseGeometry) -> bool:
        """
        Checks if the area intersects with another area.

        Args:
            area (ImxProjectArea | shapley.BaseGeometry): The other area to check, could be a shapely.Polygon.

        Returns:
            (bool): True if the area intersects with the other area, False otherwise.

        """
        other = area.polygon if isinstance(area, ImxProjectArea) else area
        return bool(self.polygon.intersects(other))



from __future__ import annotations

from enum import Enum


class AreaStatusEnum(Enum):
    """Enum to classify an IMX object based on it's geometry comparison."""

    NO_CHANGE = "NO_CHANGE"
    """Indicates that the object has not changed."""

    CREATED = "CREATED"
    """Indicates that the object has been created."""

    DELETED = "DELETED"
    """Indicates that the object is deleted."""

    MOVED = "MOVED"
    """Indicates that the object is moved from one to a other area."""

    INDETERMINATE = "INDETERMINATE"
    """Indicates that the object cant be classified."""

    def is_created_or_deleted(self):
        """
        Return True if the AreaStatus is CREATED or DELETED, False otherwise.

        Returns:
            (bool): True if the AreaStatus is CREATED or DELETED, False otherwise.
        """
        return self == AreaStatusEnum.CREATED or self == AreaStatusEnum.DELETED



from typing import Optional

from lxml import etree as ET

from imxInsights.domain.area.areaClassifier import AreaClassifier
from imxInsights.domain.area.imxProjectArea import ImxProjectArea
from imxInsights.domain.models.imxEnums import ImxSituationsEnum
from imxInsights.domain.models.imxMetadata import ProjectMetadata, ProjectNaiadeRemarks
from imxInsights.repo.imxRepo import SituationRepo
from imxInsights.utils.log import logger
from imxInsights.utils.xml_helpers import XmlFile


class ImxSituation:
    """
    A ImxSituation file contains just one situation.

    Args:
        xml_file (XmlFile): The XmlFile of the IMX file.

    Attributes:
        situation (SituationRepo): Returns the situation repo of the situation.
    """

    def __init__(self, xml_file: XmlFile, add_graph: bool = False):
        self.situation: Optional[SituationRepo] = get_imx_situation_repo(xml_file, ImxSituationsEnum.Situation)


class ImxProject:
    """
    A ImxProject contains an initial situation and project metadata, optional a new situation and the changes between those imx situations.

    Class will parse name, puic, metadata and remarks and sets the situation repos. Given the context work and user areas gml:geometry objects
    will be classified to a specific area.

    Args:
        xml_file (XmlFile): The XmlFile of the IMX file.

    Attributes:
        metadata (ProjectMetadata): Returns the situation if not project imx file.
        remarks (ProjectNaiadeRemarks): Returns the imx project if not a situation imx file.
        initial_situation (SituationRepo): Returns the situation repo of the initial situation.
        new_situation (SituationRepo): Returns the situation repo of the new situation.

    Todo:
        Move remarks to situation

    """

    def __init__(self, xml_file: XmlFile, add_graph: bool = False):
        self._element: ET.Element = xml_file.root.find(".//{*}Project")
        if self._element is not None:
            self.metadata: ProjectMetadata = ProjectMetadata(xml_file.root.find(".//{*}ProjectMetadata"))
            self.remarks: ProjectNaiadeRemarks = ProjectNaiadeRemarks(xml_file.root.find(".//{*}Remarks"))
            self.initial_situation: Optional[SituationRepo] = get_imx_situation_repo(xml_file, ImxSituationsEnum.InitialSituation, add_graph)
            self.new_situation: Optional[SituationRepo] = get_imx_situation_repo(xml_file, ImxSituationsEnum.NewSituation, add_graph)
            self._area_classifier: AreaClassifier = AreaClassifier(self.metadata.project_areas)
            if self._area_classifier is not None:
                if self.initial_situation is not None:
                    # todo: looks it can handle polygons ;)
                    self._classify_areas(self.initial_situation)
                if self.new_situation is not None:
                    self._classify_areas(self.new_situation)

    @property
    def name(self) -> str:
        """Returns the name specified on project level of an IMX file, *read only property*."""
        return str(self._element.get("name"))

    @property
    def puic(self) -> str:
        """Returns the puic specified on the project level of an IMX file, *read only property*."""
        return str(self._element.get("puic"))

    def _classify_areas(self, situation: SituationRepo) -> None:
        for entity in situation.tree.objects():
            area, classify_type = self._area_classifier.categorize_element(entity)
            if area is not None:
                assert isinstance(area.geometry, ImxProjectArea)
                entity.area = area.geometry
                entity.area_classify_type = classify_type
            else:
                entity.area = None
                entity.area_classify_type = classify_type


def get_imx_situation_repo(xml_file: XmlFile, situation: ImxSituationsEnum, add_graph: bool = False) -> Optional[SituationRepo]:
    situation_element: ET.Element = xml_file.root.find(".//{*}" + situation.value)

    if situation_element is not None:
        logger.info(f"Processing {situation.value}")
        return SituationRepo(situation_element, xml_file, add_graph)
    else:
        return None



from __future__ import annotations

from dataclasses import dataclass, field
from typing import List

from lxml import etree as ET


@dataclass
class ImxTopoObject:
    """
    Imx topo value object.

    Args:
        element (ET.Element):
        key (str):

    """

    _element: ET.Element = field(repr=False, hash=False)
    key: str = field(init=False)

    def __post_init__(self) -> None:
        ref = next((v for k, v in self._element.attrib.items() if k.endswith("Ref")), None)
        assert isinstance(ref, str), "ref key should be a string"
        assert len(ref) != 0, "ref key should not be empty"
        self.key = ref

    @property
    def element(self) -> ET.Element:
        return self._element

    @staticmethod
    def from_root_element(start: ET.Element) -> List[ImxTopoObject]:
        micro_nodes: list[ET.Element] = start.findall(".//{*}MicroNode")
        micro_links: list[ET.Element] = start.findall(".//{*}MicroLink")

        lookup = dict[ET.Element, ImxTopoObject]()
        entities: list[ET.Element] = micro_nodes + micro_links
        for entity in entities:
            lookup[entity] = ImxTopoObject(_element=entity)

        return list(lookup.values())



from typing import Protocol, TypeVar

from imxInsights.graph.imxGraphModels import GraphRoute
from imxInsights.graph.imxGraphSituationProtocol import ImxSituationRepoMinimalProtocol

T = TypeVar("T")


class ImxGraphProtocol(Protocol):
    # fmt: off
    @property
    def imx_situation(self) -> ImxSituationRepoMinimalProtocol: ...  # noqa E704

    def get_paths_between_imx_objects(self, from_object: T, to_object: T) -> list[GraphRoute]: ...  # noqa E704
    # fmt: on



from __future__ import annotations

from enum import Enum


class DiffStatusEnum(Enum):
    """Enum of possible change statuses."""

    NOT_APPLICABLE = "NOT_APPLICABLE"
    """The change does not apply."""

    NO_CHANGE = "NO_CHANGE"
    """The change is not applicable."""

    CREATED = "CREATED"
    """The change is created."""

    DELETED = "DELETED"
    """The change is deleted."""

    UPDATED = "UPDATED"
    """The change is updated."""

    UPGRADED = "UPGRADED"
    """The change is upgraded."""

    def is_changed(self) -> bool:
        """
        Returns True if the change is a created, deleted, updated or upgraded change, and False otherwise.

        Returns:
            (bool): True if the change is a created, deleted, updated or upgraded change, and False otherwise.

        """
        return self in {self.CREATED, self.DELETED, self.UPDATED, self.UPGRADED}



from __future__ import annotations

import uuid
from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional

from lxml import etree as ET

from imxInsights.utils.log import logger


class PuicKeyTypeEnum(str, Enum):
    """
    An enumeration of the different types of keys.

    Attributes:
        NORMAL_KEY: Puic attribute is used as key.
        DUMMY_KEY: Puic is a dummy value, generated new UUID4 to use as key.
        GENERATED_KEY: If no puic is given, just generated one.
        CONSTRUCTED_KEY: Does not have a puic but is relevant, we use attributes to construct a unique key.
    """

    NORMAL_KEY = "normal"
    DUMMY_KEY = "dummy_puic"
    GENERATED_KEY = "generated_key"
    CONSTRUCTED_KEY = "constructed_key"


@dataclass
class PuicConstructedByAttribute:
    """
    PuicConstructedByAttribute class that is used if a IMX object does not have a puic, it uses a set of attributes to construct an uniek key.

    Args:
        object_type (str): File path of the IMX file.
        ref_key (str): File path of the IMX file.
        attributes_to_make_unique (List[str]): File path of the IMX file.

    """

    object_type: str
    ref_key: str
    attributes_to_make_unique: List[str]

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} "
            f"ref:{self.ref_key} "
            f"attributes_to_make_unique:{' '.join(self.attributes_to_make_unique)} "
            f"object_type:{self.object_type}>"
        )


@dataclass(frozen=True)
class ImxPuic:
    """
    Unique Identifier Code, using puic if attribute is present else generate or construct a unique key.

    Args:
        valid_uuid4 (bool): returns True if the PUIC value is a valid uuid4 else False.
        puic_type (PuicConstructedByAttribute): returns the type of PUIC key.
        constructed_by_attribute (Optional[PuicConstructedByAttribute]): object that constructed PUIC by attributes.

    Raises:
        ValueError: If PUIC value is None when puic is constructed from attributes.
        AssertionError: If `constructed_puic` is None and `element` is None or `puic` attribute is not present.
        AssertionError: If `constructed_puic` is not None and `element` is None or any of the attributes in
            `constructed_puic.attributes_to_make_unique` are not present.

    """

    _puic: Optional[str | uuid.UUID] = field(default=None)
    valid_uuid4: bool = field(default=False)
    puic_type: PuicKeyTypeEnum = field(default=PuicKeyTypeEnum.NORMAL_KEY)
    constructed_by_attribute: Optional[PuicConstructedByAttribute] = field(default=None)

    def __post_init__(self):
        if self.constructed_by_attribute is not None:
            if self._puic is None:
                raise ValueError("puic value should be given when puic is constructed from attributes")
            super().__setattr__("puic_type", PuicKeyTypeEnum.CONSTRUCTED_KEY)

        if self._puic in ["00000000-0000-4000-aaad-000000000000"]:
            super().__setattr__("_puic", str(uuid.uuid4()))
            super().__setattr__("puic_type", PuicKeyTypeEnum.DUMMY_KEY)
            logger.warning(f"Dummy puic, created new puic {self._puic}")

        elif self._puic is None and self.puic_type != PuicKeyTypeEnum.CONSTRUCTED_KEY:
            super().__setattr__("_puic", str(uuid.uuid4()))
            super().__setattr__("puic_type", PuicKeyTypeEnum.GENERATED_KEY)
            logger.info(f"No Puic: create constructed key: {self._puic}")

        if isinstance(self._puic, str) and self.puic_type != PuicKeyTypeEnum.CONSTRUCTED_KEY:
            super().__setattr__("_puic", uuid.UUID(self.puic))
            super().__setattr__("valid_uuid4", True)

    def __str__(self) -> str:
        return str(self._puic)

    def __repr__(self) -> str:
        if self.puic_type == PuicKeyTypeEnum.CONSTRUCTED_KEY:
            constructed = "constructed_puic_by_attribute:" + self.constructed_by_attribute.__repr__()
        else:
            constructed = ""
        return f"<{self.__class__.__name__} {self._puic} " f"puic_type:{self.puic_type.value} " f"{constructed}>"

    @property
    def puic(self) -> str:
        """Returns the string representation of the PUIC."""
        return str(self._puic)

    @staticmethod
    def from_element(element: ET.Element, constructed_puic: Optional[PuicConstructedByAttribute] = None) -> ImxPuic:
        """Creates an instance of ImxPuic from an XML element.

        Args:
            element (bool):
            constructed_puic (Optional[PuicConstructedByAttribute]):

        Raises:
            AssertionError: If `constructed_puic` is None and `element` is None or `puic` attribute is not present.
            AssertionError: If `constructed_puic` is not None and `element` is None or any of the attributes in
                `constructed_puic.attributes_to_make_unique` are not present.

        Returns:
            (ImxPuic):

        """
        if constructed_puic is None:
            assert element is not None and "puic" in element.keys()
            return ImxPuic(element.get("puic"))

        for key in constructed_puic.attributes_to_make_unique:
            assert element is not None and key in element.keys()

        if len(constructed_puic.attributes_to_make_unique) == 0:
            custom_puic = f"{element.get(constructed_puic.ref_key)}_" + f"{constructed_puic.object_type}"
        else:
            custom_puic = f"{element.get(constructed_puic.ref_key)}_" + "".join(
                [element.get(key) for key in constructed_puic.attributes_to_make_unique]
            )

        return ImxPuic(custom_puic, constructed_by_attribute=constructed_puic)



from __future__ import annotations

from collections import defaultdict
from itertools import chain
from typing import Dict, Iterable, List

from lxml import etree as ET

from imxInsights.repo.tree.imxTopoTreeObject import ImxTopoObject
from imxInsights.repo.tree.imxTreeObject import ImxObject
from imxInsights.utils.log import logger
from imxInsights.utils.xml_helpers import NS_IMSPOOR


class ObjectTree:
    """
    Object tree of puic objects and topo objects.

    Args:
        root_element (ET.Element): root element (situation or project situation) to parse.

    """

    def __init__(self, root_element: ET.Element):
        self._root_element = root_element
        self._tree_dict: defaultdict[str, List[ImxObject]] = self._create_tree_dict(ImxObject.lookup_tree_from_root_element(root_element))
        self._update_keys()
        self._topo_dict: defaultdict[str, List[ImxTopoObject]] = self._create_topo_dict(ImxTopoObject.from_root_element(root_element))
        self._link_topo()
        self._check_junction_has_topo()

    def _update_keys(self) -> None:
        self._keys = frozenset[str]((key for key in self._tree_dict.keys()))

    @property
    def keys(self) -> frozenset[str]:
        """Get all keys of value objects in tree."""
        return self._keys

    @property
    def topo_dict(self) -> Dict[str, List[ImxTopoObject]]:
        """Get topo dict corresponding whit value objects in tree."""
        return self._topo_dict

    def objects(self) -> Iterable[ImxObject]:
        """
        Get all value objects in tree.

        Returns:
            (Iterable[ImxObject]): all value objects in tree.

        """
        return chain(*self._tree_dict.values())

    def duplicates(self) -> list[str]:
        """
        Get a list of keys that has duplicated value objects.

        Returns:
            (List[str]): list of keys.
        """
        return [k for (k, v) in self._tree_dict.items() if len(v) > 1]

    def find(self, key: str | ImxObject) -> ImxObject | None:
        """
        Find a value object given the key or the value object itself.

        Args:
            key (str | ImxObject): key or object to find in tree.

        """
        if isinstance(key, ImxObject):
            key = key.puic

        if key not in self._keys:
            return None

        match = self._tree_dict[key]
        assert len(match) == 1, f"KeyError, multiple results for {key}"
        return match[0]

    def _link_topo(self) -> None:
        for topo_object in self._topo_dict.values():
            puic_object = self.find(topo_object[0].key)
            if puic_object is not None:
                puic_object.register_topo_object(topo_object[0])
            else:
                logger.error(f"Topo object {topo_object[0].key} should have a puic object")

    def _check_junction_has_topo(self) -> None:
        for puic_object in self._tree_dict.values():
            parent = puic_object[0].element.getparent()
            if parent.tag == f"{{{NS_IMSPOOR}}}Junctions":
                try:
                    self._topo_dict[puic_object[0].puic]
                except ValueError:
                    logger.error(f"Junction object {puic_object[0].puic} should have a topo object")

    @staticmethod
    def _create_tree_dict(objects: Iterable[ImxObject]) -> defaultdict[str, list[ImxObject]]:
        result = defaultdict[str, list[ImxObject]](list)
        for o in objects:
            result[o.puic].append(o)
        duplicated = [o for o in objects if len(result[o.puic]) != 1]
        assert len(duplicated) == 0, f"KeyError, multiple results for {[item.puic for item in duplicated]}"
        return result

    @staticmethod
    def _create_topo_dict(topo_objects: Iterable[ImxTopoObject]) -> defaultdict[str, list[ImxTopoObject]]:
        result = defaultdict[str, list[ImxTopoObject]](list)
        for o in topo_objects:
            result[o.key].append(o)

        duplicated = [o for o in topo_objects if len(result[o.key]) != 1]
        assert len(duplicated) == 0, f"KeyError, multiple results for {[item.key for item in duplicated]}"
        return result



from __future__ import annotations

from dataclasses import dataclass, field
from typing import List, Optional

from lxml import etree as ET
from shapely import GeometryCollection
from shapely.geometry.base import BaseGeometry
from shapelyM.measureLineString import MeasureProfile


@dataclass
class RailConnectionInfos:
    """
    Dataclass to list to list all RailConnectionInfo nodes of a value object.

    Args:
        rail_infos (List[RailConnectionInfoPoint | RailConnectionInfoLine | RailConnectionInfo]): list of all railConnectionInfo's.

    """

    rail_infos: List[RailConnectionInfoPoint | RailConnectionInfoLine | RailConnectionInfo] = field(default_factory=list)

    @staticmethod
    def from_element(element: ET.Element, object_type: str) -> Optional[RailConnectionInfos]:
        """Rail connection info factory."""
        rail_connection_infos_elements = element.findall(f".//{{*}}{object_type}")
        if len(rail_connection_infos_elements) == 0:
            return None

        _out = RailConnectionInfos()
        for rail_connection_info_element in rail_connection_infos_elements:
            ref = rail_connection_info_element.get("railConnectionRef")
            at_measure = rail_connection_info_element.get("atMeasure")
            from_measure = rail_connection_info_element.get("fromMeasure")
            to_measure = rail_connection_info_element.get("toMeasure")
            direction = rail_connection_info_element.get("direction")

            if ref and at_measure and direction:
                _out.rail_infos.append(RailConnectionInfoPoint(ref, float(at_measure), direction))
            elif ref and from_measure and to_measure:
                _out.rail_infos.append(RailConnectionInfoLine(ref, float(from_measure), float(to_measure), direction))
            else:
                _out.rail_infos.append(RailConnectionInfo(ref))

        return _out

    def __iter__(self):
        for rail_info in self.rail_infos:
            yield rail_info

    def get_geometry_collection(self):
        return GeometryCollection([item.geometry for item in self.rail_infos])

    # todo: make methode to check if invalid relations


@dataclass
class RailConnectionInfo:
    """
    RailConnectionInfo base object.

    Geometry and project will be set by repo.

    Args:
        ref (str): unique key (puic) to railConnection
        projection (Optional[BaseGeometry]): project if point else None.

    """

    ref: str
    present_in_dataset: bool = field(init=False, default=True)
    geometry: Optional[BaseGeometry] = field(init=False, default=None)
    _projection: Optional[MeasureProfile] = field(init=False, default="Not Implemented")


@dataclass
class RailConnectionInfoPoint(RailConnectionInfo):
    """
    RailConnectionInfo Point extension.

    Args:
        at_measure (float): at measure along railConnection.
        direction (): direction along railConnection.
    """

    at_measure: float
    direction: str
    at_measure_3d: Optional[float] = field(init=False, default=None)

    # todo: create projection property, set "point on track" and calculate 3d_measure
    # todo: create side_of_track property


@dataclass
class RailConnectionInfoLine(RailConnectionInfo):
    """
    RailConnectionInfo Line extension.

    A to measure can be lower than a from measure and vice verse.

    Args:
        from_measure (float): from measure along railConnection.
        to_measure (float): from measure along railConnection.
        direction (): direction along railConnection.
    """

    from_measure: float
    to_measure: float
    direction: str

    # todo: create geometry property, set line



from dataclasses import dataclass, field
from typing import Optional

from shapely import GeometryCollection, LineString
from shapely.geometry.base import BaseGeometry
from shapelyM import MeasureLineString


@dataclass
class ImxGeometry:
    """
    A class representing an ImxGeometry.

    Args:
        shapely (shapely.BaseGeometry): shapely geometry of the value object.

    Attributes:
        azimuth (Optional[float]): the azimuth (rotation from north in direction east) of a value object, if not present None
        shapelyM (Optional[MeasureLineString]): shapely Measure linestring if shapely object is linestring else None.

    """

    shapely: BaseGeometry = GeometryCollection()
    azimuth: Optional[float] = None
    shapelyM: Optional[MeasureLineString] = field(init=False)

    def __post_init__(self) -> None:
        if isinstance(self.shapely, LineString):
            self.shapelyM = MeasureLineString(list(self.shapely.coords))
        else:
            self.shapelyM = None



from collections import OrderedDict
from datetime import datetime, timezone
from typing import Dict, List, Optional

import numpy as np
import pandas as pd
from shapely.geometry.base import BaseGeometry

from imxInsights import DiffStatusEnum
from imxInsights import __version__ as libary_version
from imxInsights.diff.compair import ImxObjectCompare
from imxInsights.repo.imxRepo import SituationRepo
from imxInsights.report.file_info import (
    sort_pandas_dataframe_columns,
    write_situation_info,
)
from imxInsights.report.refDisplay import ImxRefDisplay
from imxInsights.utils.log import logger
from imxInsights.utils.shapely_geojson import GeoJsonFeature, GeoJsonFeatureCollection
from imxInsights.utils.shapely_helpers import ShapelyTransform


class ImxDiff:
    """
    Calculated difference between two imx situations.

    Args:
        situation_repo_1 (SituationRepo):
        situation_repo_2 (SituationRepo):

    """

    def __init__(self, situation_repo_1: SituationRepo, situation_repo_2: SituationRepo):
        self.situation_repo_1 = situation_repo_1
        self.situation_repo_2 = situation_repo_2
        self._diff = list(ImxObjectCompare.object_tree_factory(self.situation_repo_1.tree, self.situation_repo_2.tree))
        self.added_color = "#ff0000"
        self.deleted_color = "#002aff"
        self.change_color = "#3deb34"
        self.no_change_color = "#d9d9d9"

    def _filter_by_types_or_path(self, object_type_or_path: str) -> List[ImxObjectCompare]:
        # TODO: make object_type_or_path a list input to filter sets of types
        search = "path" if "." in object_type_or_path else "tag"
        return [item for item in self._diff if item.__getattribute__(search) == object_type_or_path]

    @staticmethod
    def _create_record_dict(item, geometry: Optional[bool] = False, compact_view: Optional[bool] = True) -> Dict:
        def _get_areas(item: ImxObjectCompare):
            area_a = ""
            area_b = ""

            if item.area_status.name == "INDETERMINATE":
                area_a = "INDETERMINATE"
                area_b = "INDETERMINATE"

            else:
                if item.a is not None:
                    area_a = item.a.area.name
                if item.b is not None:
                    area_b = item.b.area.name
            return area_a, area_b

        def _get_parent(item: ImxObjectCompare):
            parent = ""
            if hasattr(item.a, "parent") and hasattr(item.b, "parent"):
                if item.a.parent is not None and item.b.parent is not None:
                    if item.a.parent.puic != item.b.parent.puic:
                        parent = f"{item.a.parent.puic}->{item.b.parent.puic}"
                    else:
                        parent = f"{item.a.parent.puic}"
            elif hasattr(item.a, "parent") and item.a.parent is not None:
                parent = f"{item.a.parent.puic}"
            elif hasattr(item.b, "parent") and item.b.parent is not None:
                parent = f"{item.b.parent.puic}"

            return parent

        def _get_km(item):
            KM_dict = {}
            if geometry is not None:
                if item.a is not None and item.a.km_values is not None:
                    _ = item.a.km_values
                    KM_dict["lr_km_values_a"] = _[0] if len(_) == 1 else "\n".join(_)
                else:
                    KM_dict["lr_km_values_a"] = None

                if item.b is not None and item.b.km_values is not None:
                    _ = item.b.km_values
                    KM_dict["lr_km_values_b"] = _[0] if len(_) == 1 else "\n".join(_)
                else:
                    KM_dict["lr_km_values_b"] = None
            return KM_dict

        area_a, area_b = _get_areas(item)
        parent = _get_parent(item)
        km_dict = _get_km(item)

        props = {
            "puic": item.puic,
            "path": item.path,
            "tag": item.tag,
            "area_a": area_a,
            "area_b": area_b,
            "area_status": item.area_status.name,
            "diff_status": item.diff_status.name,
        }

        imx_props = item.changes.to_dict(compact=compact_view)

        if "@puic" in imx_props.keys():
            del imx_props["@puic"]

        return props | imx_props | {"parent": parent} | km_dict  # | geometry_dict

    def _get_all_as_record_dict_path_is_key(self, geometry: Optional[bool] = False) -> OrderedDict[str, List[Dict]]:
        record_dict = OrderedDict()

        for item in self._diff:
            record = self._create_record_dict(item, geometry)

            if item.path in record_dict.keys():
                record_dict[item.path].append(record)
            else:
                record_dict[item.path] = [record]

        return record_dict

    @staticmethod
    def _get_dataframe_from_records(records: List[Dict]):
        df = pd.DataFrame.from_records(records)
        df["index"] = df["puic"]
        df.set_index("index", inplace=True)
        df = df.fillna("")
        return df

    def get_by_status(self, status: Optional[List[DiffStatusEnum]] = None, object_type_or_path: Optional[str] = None) -> List[ImxObjectCompare]:
        data = self._filter_by_types_or_path(object_type_or_path) if object_type_or_path else self._diff
        return [item for item in data if item.diff_status in status]

    def get_by_puic(self, puic: str) -> ImxObjectCompare:
        return [item for item in self._diff if item.puic == puic][0]

    def get_by_type(self, imx_types: List[str]) -> List[ImxObjectCompare]:
        return [item for item in self._diff if item.tag in imx_types]

    def get_by_path(self, paths: List[str]) -> List[ImxObjectCompare]:
        return [item for item in self._diff if item.path in paths]

    @classmethod
    def _add_zero_index(cls, df):
        # todo: move to utils and use in population view of situation as well
        # Check if there's a column containing ".0."
        if any(".0." in path for path in df.columns):
            # Extract the column without ".0."
            columns_with_zero = [_ for _ in list(df.columns) if ".0." in _]
            # replace the .0. and find the non zero'ed colum
            columns_without_zero = [_.replace(".0.", ".") for _ in columns_with_zero]
            columns_to_change = [_ for _ in df.columns if _ in columns_without_zero]
            column_mapping = list(zip(columns_to_change, columns_with_zero))

            for item in column_mapping:
                df[item[1]] = df[item[1]].replace("", np.nan)
                df[item[1]] = df[item[1]].fillna(df[item[0]])
                df[item[1]] = df[item[1]].replace(np.nan, "")
                del df[item[0]]

        return df

    def pandas_dataframe_dict(self, geometry: Optional[bool] = False) -> Dict[str, pd.DataFrame]:
        """
        Returns a dictionary of all difference as pandas dataframe, key is path of imx object.

        Args:
            geometry: boolean if True include wkt_hex.

        Returns
            dict of all object pandas dataframes.
        """
        dataframe_dict = {key: self._get_dataframe_from_records(value) for key, value in self._get_all_as_record_dict_path_is_key(geometry).items()}
        return {key: self._add_zero_index(value) for key, value in dataframe_dict.items()}

    def pandas_dataframe(self, object_type_or_path: str = None, geometry: Optional[bool] = False) -> pd.DataFrame:
        """
        Returns the differences as a pandas DataFrame.

        Args:
            object_type_or_path (str): The object type or path to return the differences, defaults to None .
            geometry (Optional[bool]): Whether to include the shapely geometry in the DataFrame, defaults to False.

        Returns:
            (pd.DataFrame): A pandas DataFrame representing the filtered differences.
        """
        dataframe = self._get_dataframe_from_records(
            [self._create_record_dict(item, geometry) for item in self._filter_by_types_or_path(object_type_or_path)]
        )
        return self._add_zero_index(dataframe)

    def generate_excel(self, file_path: str, colors: bool = True, ref_display: bool = True, parent_path: bool = True, overview: bool = True) -> None:
        """
        Generates an Excel file that highlights the differences between datasets.

        Each difference is stylized according to its nature (e.g., created, updated, deleted) and can be further
        annotated with references and parent paths if specified.

        This function utilizes formatting to visualize these differences within the Excel file, supporting better
        understanding and analysis of the changes. The coloring scheme (green for updates, red for creations,
        blue for deletions, and gray for no changes), sorts and formats the data according to the specified conditions,
        and then generates an Excel workbook with multiple sheets, each corresponding to different segments
        of the dataset being compared. Each sheet includes a detailed breakdown of changes, with optional reference
        and parent path information, and is formatted for clarity and ease of analysis.

        Args:
            file_path (str): The output file path where the Excel file will be saved.
            colors (bool, optional): If True, applies conditional coloring to highlight different types of
                differences. Defaults to True.
            ref_display (bool, optional): If True, includes reference display information within the Excel
                sheet, enhancing contextual understanding of the differences. Defaults to True.
            parent_path (bool, optional): If True, includes the parent path for each item, providing additional
                hierarchical context. Defaults to True.
            overview (bool, optional): If True, includes a metadata overview sheet.

        """
        sort_status_list = ["CREATED", "UPDATED", "UPGRADED", "NO_CHANGE", "DELETED"]

        def format_diff_status(val):
            if val in ["CREATED"]:
                return f"border: 5px solid {self.added_color}"
            elif val in ["UPDATED", "UPGRADED"]:
                return f"border: 5px solid {self.change_color}"
            elif val in ["DELETED"]:
                return f"border: 3px solid {self.deleted_color}"
            return ""

        def format_moved(val):
            if val in ["MOVED"]:
                return "background-color: #e0e33c"
            return ""

        def format_diff_value(val):
            if isinstance(val, str):
                if val[:3] == "++ ":
                    return f"border: 3px solid {self.added_color}"
                elif " -> " in val:
                    return f"border: 3px solid {self.change_color}"
                elif val[:3] == "-- ":
                    return f"border: 2px solid {self.deleted_color}"
                else:
                    return f"background-color: {self.no_change_color}"
            return ""

        def gray_out_no_change_row(row):
            if row["diff_status"] in ["NO_CHANGE"]:
                return [f"background-color: {self.no_change_color}"] * len(row)
            return [""] * len(row)

        def center_values(val):
            return "text-align: center"

        logger.info("generate Excel diff")
        df_dict = self.pandas_dataframe_dict()

        writer = pd.ExcelWriter(file_path, engine="xlsxwriter")
        workbook = writer.book

        worksheet_info = workbook.add_worksheet("info")

        worksheet_info.set_column(0, 0, 25)
        worksheet_info.set_column(1, 1, 150)

        worksheet_info.write(0, 0, "info")
        worksheet_info.write(1, 0, "process datestamp")
        worksheet_info.write(1, 1, datetime.now().astimezone(timezone.utc).strftime("%Y-%m-%d %H:%M:%S.%f"))
        worksheet_info.write(2, 0, "imxInsight version")
        worksheet_info.write(2, 1, libary_version)

        write_situation_info(worksheet_info, self.situation_repo_1, 4, "a")
        write_situation_info(worksheet_info, self.situation_repo_2, 11, "b")

        worksheet_info.set_tab_color("#6699ff")

        #####################################
        # todo: refactor below to make more readable
        def create_puic_path(df: pd.DataFrame):

            def construct_parent_path(row):
                parent = row["parent"]
                if parent:
                    if "->" in parent:
                        raise ValueError("The parent has changed. This is not supported and was not intended as a concept.")

                    parent_row = df[df["puic"] == parent].iloc[0]
                    temp = construct_parent_path(parent_row) + "." + parent
                    if temp[0] == ".":
                        return temp[1:]
                    return temp
                else:
                    return ""

            def construct_index(row):
                if "." in row["path"]:
                    path_first = row["path"].split(".")[0]
                    path_last = row["path"].split(".")[-1]
                else:
                    path_first = row["path"]
                    path_last = row["path"]

                if path_first == path_last:
                    return path_first + row["parent_path"] + "." + row["puic"] + ".00000." + row["diff_status"]

                return path_first + "." + row["parent_path"] + "." + row["puic"] + ".00000." + path_last + ".00000." + row["diff_status"]

            # Apply the function to each row
            if parent_path:
                df["parent_path"] = df.apply(construct_parent_path, axis=1)
                df["new_index"] = df.apply(construct_index, axis=1)

            max_levels = df["path"].str.count(r"\.") + 1
            max_levels = max_levels.max()
            split_columns = [f"lvl{i + 1}" for i in range(max_levels)]
            df[split_columns] = df["path"].str.split(".", expand=True)

            if parent_path:
                sorted_df = df.sort_values(by=["new_index"])
                sorted_df.set_index("new_index", inplace=False)
                sorted_df.reset_index(drop=True, inplace=True)

                sorted_df = sorted_df[
                    ["new_index", "path", "tag", "puic", "area_status", "diff_status"]
                    + split_columns
                    + ["Metadata.@lifeCycleStatus", "Metadata.@source", "Metadata.@isInService", "Metadata.@registrationTime", "@name"]
                ]
            else:
                df.reset_index(drop=True, inplace=True)
                sorted_df = df

            return sorted_df, split_columns

        basic_columns = ["path", "tag", "puic", "parent", "area_a", "area_b", "area_status", "diff_status", "@name"]
        metadata_columns = list(set([col for key, df in df_dict.items() for col in df.columns if col.startswith("Metadata")]))
        columns_to_keep = basic_columns + [col for col in metadata_columns if col not in basic_columns]
        dfs_with_selected_columns = {
            key: df[[col for col in columns_to_keep if col in columns_to_keep and col in df.columns]] for key, df in df_dict.items()
        }
        concatenated_df = pd.concat(dfs_with_selected_columns.values())

        concatenated_df, split_columns = create_puic_path(concatenated_df)

        if overview:
            overview_sheet = workbook.add_worksheet("meta-overview")

            tree_not_visible_style = workbook.add_format({"font_color": "#FFFFFF"})
            created_style = workbook.add_format({"font_color": "red", "bold": True})
            deleted_style = workbook.add_format({"font_color": "blue", "bold": True, "font_strikeout": True})
            changed_style = workbook.add_format({"font_color": "green", "bold": True})
            error_style = workbook.add_format({"font_color": "purple", "bold": True})

            first_after_levels = 5 + len(split_columns)

            overview_sheet.write(0, 0, "index")
            overview_sheet.write(0, 1, "path")
            overview_sheet.write(0, 2, "tag")
            overview_sheet.write(0, 3, "puic")
            overview_sheet.write(0, 4, "area_status")
            overview_sheet.write(0, 5, "diff_status")
            for i in range(6, (6 + len(split_columns))):
                overview_sheet.write(0, i, f"lvl_{i - len(split_columns)}")

            overview_sheet.write(0, first_after_levels + 1, "@name")
            overview_sheet.write(0, first_after_levels + 2, "Metadata.@lifeCycleStatus")
            overview_sheet.write(0, first_after_levels + 3, "Metadata.@source")
            overview_sheet.write(0, first_after_levels + 4, "Metadata.@isInService")
            overview_sheet.write(0, first_after_levels + 5, "Metadata.@registrationTime")

            for index, row in concatenated_df.iterrows():
                index = index + 1
                status = row["diff_status"]
                if status == "NO_CHANGE":
                    cell_format = None
                elif status == "CREATED":
                    cell_format = created_style
                elif status == "DELETED":
                    cell_format = deleted_style
                elif status in ["UPDATED", "UPGRADED"]:
                    cell_format = changed_style
                else:
                    cell_format = error_style

                overview_sheet.write(index, 0, index)
                overview_sheet.write(index, 1, row["path"])
                overview_sheet.write(index, 2, row["tag"])
                overview_sheet.write(index, 3, row["puic"])
                overview_sheet.write(index, 4, row["area_status"])
                overview_sheet.write(index, 5, row["diff_status"], cell_format)

                overview_sheet.write(index, first_after_levels + 1, str(row.get("@name")))
                overview_sheet.write(index, first_after_levels + 2, str(row.get("Metadata.@lifeCycleStatus", "")))
                overview_sheet.write(index, first_after_levels + 3, str(row.get("Metadata.@source", "")))
                overview_sheet.write(index, first_after_levels + 4, str(row.get("Metadata.@isInService", "")))
                overview_sheet.write(index, first_after_levels + 5, str(row.get("Metadata.@registrationTime", "")))

                for idx, value in enumerate(split_columns, start=1):
                    if idx < len(split_columns):
                        if row[split_columns[idx]] is None:
                            overview_sheet.write(index, 5 + idx, row[value], cell_format)
                        else:
                            overview_sheet.write(index, 5 + idx, row[value], tree_not_visible_style)
                    else:
                        overview_sheet.write(index, 5 + idx, row[value], cell_format)

            overview_sheet.autofit()

            for i in range(6, (6 + len(split_columns))):
                if i == (6 + len(split_columns) - 1):
                    overview_sheet.set_column(i, i, 50)
                else:
                    overview_sheet.set_column(i, i, 3)

            overview_sheet.set_column(1, 2, options={"level": 1, "collapsed": True})
            overview_sheet.set_tab_color("#6699ff")

        #####################################

        puic_ref_display = ImxRefDisplay(self.situation_repo_1, self.situation_repo_2)
        wrapped_text = workbook.add_format({"text_wrap": True})

        for key, df in sorted(df_dict.items()):
            logger.info(f"generating sheet {key}")
            # first column order
            column_order_list = ["puic", "path", "tag", "parent", "area_a", "area_b", "area_status", "diff_status", "@name"]
            df = sort_pandas_dataframe_columns(df, column_order_list)

            # sort on diff status
            df["diff_status"] = pd.Categorical(df["diff_status"], categories=sort_status_list, ordered=True)
            df = df.sort_values("diff_status")
            df = df.reset_index(drop=True)

            # make color book 😂
            if colors:
                styler = (
                    df.style.map(format_diff_status, subset=["diff_status"])
                    .map(format_moved, subset=["area_status"])
                    .map(format_diff_value)
                    .apply(gray_out_no_change_row, axis=1)
                    .map(center_values, subset=["area_a", "area_b", "area_status", "diff_status"])
                    .set_properties(**{"font-size": "10pt"})
                )
            else:
                styler = df

            if ref_display:
                exclude_columns = ["puic", "parent"]
                df.loc[:, ~df.columns.isin(exclude_columns)] = df.loc[:, ~df.columns.isin(exclude_columns)].map(puic_ref_display.get_display)

            if parent_path:
                df["parent"] = df["parent"].map(puic_ref_display.get_parent_display)

            shorten_sheet_name = f"{key[:14]}...{key[-14:]}" if len(key) > 30 else key
            styler.to_excel(excel_writer=writer, sheet_name=shorten_sheet_name)
            worksheet = writer.sheets[shorten_sheet_name]

            # gray out sheets with no changes
            if list(set(df["diff_status"])) == ["NO_CHANGE"]:
                worksheet.set_tab_color(self.no_change_color)

            # Get the dimensions of the dataframe.
            (max_row, max_col) = df.shape

            worksheet.set_column(0, max_col - 1, None, cell_format=wrapped_text)

            # Set the auto filter.
            worksheet.autofilter(0, 0, max_row, max_col)

            # autofit and freeze header
            worksheet.autofit()
            worksheet.freeze_panes(1, 0)
            if colors:
                worksheet.protect("", {"autofilter": True, "format_cells": True, "objects": True})

        logger.info("save Excel diff as .xlsx file")
        writer.close()

    def as_geojson(self, object_type_or_path: str) -> GeoJsonFeatureCollection:
        """
        Returns a GeoJSON of the differences.

        Args:
            object_type_or_path (str): The object type or path to filter the differences.

        Returns:
            (GeoJsonFeatureCollection): A GeoJsonFeatureCollection representing the filtered differences.
        """
        features = []
        for item in self._filter_by_types_or_path(object_type_or_path):
            difference_dict = item.changes.to_dict()

            if item.b is not None and item.b.km_values:
                km_value_b = item.b.km_values[0] if len(item.b.km_values) == 1 else ", ".join(item.b.km_values) if len(item.b.km_values) != 0 else ""
            else:
                km_value_b = None

            if item.diff_status.value == "DELETED":
                if item.a.km_values:
                    km_value_a = (
                        item.a.km_values[0] if len(item.a.km_values) == 1 else ", ".join(item.a.km_values) if len(item.a.km_values) != 0 else ""
                    )
                else:
                    km_value_a = None
                features.append(
                    GeoJsonFeature(
                        geometry_list=[ShapelyTransform.rd_to_wgs(item.a.shapely if item.a.shapely is not None else BaseGeometry())],
                        properties=difference_dict
                        | {"diff_status": item.diff_status.value, "color": self.deleted_color, "km_a": km_value_a, "path": item.path},
                    )
                )
            elif item.diff_status.value == "CREATED":
                features.append(
                    GeoJsonFeature(
                        geometry_list=[ShapelyTransform.rd_to_wgs(item.b.shapely if item.b.shapely is not None else BaseGeometry())],
                        properties=difference_dict
                        | {"diff_status": item.diff_status.value, "color": self.added_color, "km_b": km_value_b, "path": item.path},
                    )
                )
            elif item.diff_status.value == "UPDATED" or item.diff_status.value == "UPGRADED":
                features.append(
                    GeoJsonFeature(
                        geometry_list=[ShapelyTransform.rd_to_wgs(item.b.shapely if item.b.shapely is not None else BaseGeometry())],
                        properties=difference_dict
                        | {"diff_status": item.diff_status.value, "color": self.change_color, "km_b": km_value_b, "path": item.path},
                    )
                )

            else:
                features.append(
                    GeoJsonFeature(
                        geometry_list=[ShapelyTransform.rd_to_wgs(item.b.shapely if item.b.shapely is not None else BaseGeometry())],
                        properties=difference_dict
                        | {"diff_status": item.diff_status.value, "color": self.no_change_color, "km_b": km_value_b, "path": item.path},
                    )
                )
        return GeoJsonFeatureCollection(geojson_features=features)

    def _get_all_paths(self):
        return list(set([item.path for item in self._diff]))

    def _get_all_types(self):
        return list(set([item.tag for item in self._diff]))

    def generate_geojson_dict(self, key_based_on_type: bool = False) -> Dict[str, GeoJsonFeatureCollection]:
        """
        Generates all GeoJSONs of the differences.

        Args:
            key_based_on_type (bool): Whether to use object type as the key in the dictionary, defaults to False.

        Returns:
            (Dict[str, GeoJsonFeatureCollection]): A dictionary of GeoJsonFeatureCollections representing the differences.
        """
        out_dict = {}
        if key_based_on_type:
            for imx_type in self._get_all_types():
                out_dict[imx_type] = self.as_geojson(imx_type)

        for imx_path in self._get_all_paths():
            out_dict[imx_path] = self.as_geojson(imx_path)

        return out_dict
